/* ****************************************** */
/* Emu71: HP-71B software emulator for PC     */
/*                                            */
/* emu71.c   main module - PC version         */
/*                                            */
/* Version 2.45, J-F Garnier 1996-2016        */
/* ****************************************** */


/* Copyright (c) 1996-2016  J-F Garnier
'
' This program is free software; you can redistribute it and/or
' modify it under the terms of the GNU General Public License
' as published by the Free Software Foundation; either version 2
' of the License, or (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with this program; if not, write to the Free Software
' Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <alloc.h>
#include <time.h>



/* definition of UNIT_TICKS for clock() units */
#if __TURBOC__  >= 0x400
#define UNIT_TICKS 1
#else
#define UNIT_TICKS 55
#endif

extern unsigned _stklen = 8000;


#include "saturn.h"

#include "version.h"

extern int commande(void);  /* monit.c */
extern long lastaddr;	   

extern int ilmailbox(int n, char *mb);

extern void update_display(char *buf1, char *buf2, int cpos, int ctype);


char ROM71[40]= "rom71.bin";

#define RAMSIZE 32

unsigned char buffer[2048];    /* temporary buffer  */

char f_trace;   /* trace */

int fmonit;

/* variables module description  */
struct portdevice {
  int nuport;      /* port */
  int type;        /* module type : A=RAM, O=ROM, R=HRD, E=EEPROM */
  int taille;      /* file size */
  char nomfic[40]; /* file associated with the module */
  long adresse;    /* configuration address */
  int configure;   /* configuration status */
  long id;         /* identifier (reponse to c=id) */
  int seg;	   /* segment holding the module */
} ;


struct portdevice device71[16]; /* 16 devices 71 */


int speed_emu;  /* number of instructions emulated per ticks 71 (1/512 s) */

/* variables simulation timer */
long last_timer=0;

/* for module desas.c : */
char from=1; /* changed JFG 17/05/07 */

char f_xil=0;

int nbdev;  /* number of devices */
int tabdev[32];

int flowpower;

#if __TURBOC__ < 0x200
void _fmemcpy(int far *dest, int far *src, int n)
{
  n /= 2;
  while (n--) *dest++ = *src++;
}
void _fmemset(int far *dest, int val, int n)
{
  n /= 2;
  while (n--) *dest++ = val;
}
#endif



/* macro to build a pointer into Emu71 memory space. Assume a is even */
#define MK_FP71(a) MK_FP(tab_seg[a>>16],(int)((a>>1)&0x7fff))



/* ************************************************************* */
/* keyboard management                                           */
/* Emu71 emulates the keyboard at key buffer level,
/* not at register level.
/* This makes the emulation faster, but also made the emulation
/* of the RAM debugger more complex, since the debugger uses a
/* different key buffer                                          */
/* ************************************************************* */

  /*
    2f441 atndis  1     attn disable
    2f442 atnflg  1     attn key hit flag
    2f443 keyptr  1     key buffer pointer
    2f444 keybuf  15*2  key buffer
  */
  /* for debugger:
    2f43c: VECTOR 5      external interrupt handler vector
    VECTOR should be x0159
    if yes
      x08b2 MANHOL
      x08d0 DATNFL
      x08d1 DKYPTR
      x08d2 DKYBUF
  */




/* ****************************************** */
/* keybufst()                                 */
/*                                            */
/* read the attn flag of the key buffer       */
/* ****************************************** */
int keybufst(void)
{
  unsigned char far *ptr;
  unsigned char far *ptr1;
  int n;

  ptr =MK_FP71(0x2f442L);          /* attn flag */
  ptr1=MK_FP71(0x2f43cL);          /* VECTOR */

  if ( ((*ptr1)==0x59) && (*(ptr1+1)==0x01) ) {
    /* RAM debugger */
    n=(*(ptr1+2))&15;
    if (tab_seg[n]!=0) {
      ptr1=MK_FP(tab_seg[n],0x0459);  /* test MANHOL: use KEYBUF or DKYBUF ? */
      if (((*ptr1)&0x1)==0) {
        ptr=MK_FP(tab_seg[n],0x0468);  /* DATNFL: debugger attn flag */
      }
    }
  }

  n = *ptr;
  return(n>>4);
}



/* ****************************************** */
/* push_key()                                 */
/*                                            */
/* push a key into the key buffer              */
/* ****************************************** */
int push_key(int key)
{

  unsigned char far *ptr;
  unsigned char far *ptr1;
  int n, fdbg;

  ptr =MK_FP71(0x2f442L);  /* ATNFLG: attn flag */
  ptr1=MK_FP71(0x2f43cL);  /* VECTOR */

  fdbg=0;
  if ( ((*ptr1)==0x59) && (*(ptr1+1)==0x01) ) {
    /* RAM debugger */
    n=(*(ptr1+2))&15;
    if (tab_seg[n]!=0) {
      ptr1=MK_FP(tab_seg[n],0x0459);  /* test MANHOL: use KEYBUF or DKYBUF ? */
      if (((*ptr1)&0x1)==0) {
        ptr=MK_FP(tab_seg[n],0x0468);  /* DATNFL: debugger attn flag */
        fdbg=1;
      }
    }
  }
  else if ( ((*ptr1)==0x87) && (*(ptr1+1)==0xf9) ) {
    n=(*(ptr1+2))&15;
    if (n==2) {
      /* VECTOR points to 2f987, this is the HP ONKEYLX */
      /* set flag Except (12) */
      status |= 1<<12;
    }
  }

  n = *ptr;
  if ( (key==43) && !((*(ptr-1))&0xf0) ) {
    /* ATTN key: reset buffer, and decrement attn key hit flag */
    n &= 0x0f;
    if (n!=1) n--;
    n &= 0x0f;
    if (!fdbg) {
      /* set flag Except (12) */
      status |= 1<<12;
    }
  }
  /* add key into buffer */
  if ((n&0xf0) < 0xf0) {
    n += 0x10;
    *ptr=n;
    n >>=4;
    ptr += n;
    *ptr=key;
    return(0);
  }
  else
    return(1);
}



/* ****************************************** */
/* do_recover()                               */
/*                                            */
/* manage the recover (ON-/) sequence         */
/* of the RAM debugger                        */
/* ****************************************** */
void do_recover(void)
{
  unsigned char far *ptr1;
  int n, m;
  unsigned long t;

  /* fill in the register save area (for dbg) */
  ptr1=MK_FP71(0x2f400L);  /* INTR4 */

  t  =  regR4[0] & 0xfff00000L  ;
  t |= (regD0    &    0xfffffL );
  _fmemcpy(ptr1,&t,4); ptr1 +=4;  /* save R4[7-5] and D0 in INTR4 */
  _fmemcpy(ptr1,&regR4[1],4); ptr1 +=4; /* save R4[15-8] */
  _fmemcpy(ptr1,regA,8); ptr1 +=8;      /* save A */
  _fmemcpy(ptr1,regB,8); ptr1 +=8;      /* save B */

  memcpy(regR4,regC,8);    /* save C in R4 */

  if (flagdec) t=10; else t=0;
  t<<=4;
  t+=Pointer;
  t<<=4;
  if (Carry==0) t |= 1;
  t<<=20;
  t |= regPC & 0xfffffL;
  _fmemcpy(ptr1,&t,4);      /* misc save  */


  ptr1=MK_FP71(0x2f43cL);  /* VECTOR */

  regPC=0xff3;  /* WARMST */
  /* is VECTOR a call to dbg? */
  if ( ((*ptr1)==0x59) && (*(ptr1+1)==0x01) ) {
    /* yes. */
    n=(*(ptr1+2))&15;
    if (tab_seg[n]!=0) {
      /* save D1 in x08AD */
      ptr1=MK_FP(tab_seg[n],0x0456);
      m= (*ptr1) & 0x0f;
      *ptr1++ = m | (regD1<<4);
      *ptr1++ = (regD1 >> 4);
      *ptr1 = (regD1 >> 12);
      /* clear one byte at x08FE */
      ptr1=MK_FP(tab_seg[n],0x047F);
      *ptr1 =0;
      /* clear bit 0 of 2e102 */
      ptr1=MK_FP(tab_seg[2],0x7081);
      *ptr1 &= 0xfe;

      /* set PC to enter debugger */;
      regPC=n;
      regPC <<= 16;
      regPC+=0x24d;  /* RECOVER */

    }
  }
}



/* ****************************************** */
/* traite_touche()                            */
/*                                            */
/* manage the keyboard                        */
/* flag fwait means: wait for a key           */
/* ****************************************** */
int static traite_touche(int fwait)
{
  int k;

    k=readkey(fwait);
    if ((k>0)&&(k<=168)) {
      push_key(k);
    }

  return(k);
}



/* ************************************************************* */
/* display management                                            */
/* Emu71 emulates the LCD at display buffer level,
/* not at LCD bit-map level.
/* This makes the emulation faster, but also made the emulation
/* of the RAM debugger more complex, since the debugger uses a
/* different display buffer                                      */
/* ************************************************************* */

/*
 DSPSTA (2f475) Display status
	(2f478)
	(2f479) 2: cursor on(0)/off(1), 3: insert(1)/replace(0)
 FIRSTC (2f47c) Buffer position of 1st char.
 CURSOR (2f47e) Buffer position of cur.
 DSPBFS (2f480) 96 char.
*/



/* ****************************************** */
/* traite_display()                           */
/*                                            */
/* manage the LCD display                     */
/* flag frun means: CPU running ('*' announc) */
/* ****************************************** */
void static traite_display(int frun)
{
  static char derligne[98] = "";
  static char derflags[8] = "";
  static int derpos=-1;
  static int deretat=-1;
  static int derdeb=0;
  char ligne[98];
  char flags[8];
  char far *ptr;
  char far *ptr1;
  int xp, xs, deb;
  int faff, fcsr, flg, fdbg, n;

  ptr =MK_FP71(0x2f480L);  /* DSPBFS mainframe display buffer */
  ptr1=MK_FP71(0x2f43cL);  /* VECTOR */

  /* check which buffer to use */
  fdbg=0;
  if ( ((*ptr1)==0x59) && (*(ptr1+1)==0x01) ) {
    n=(*(ptr1+2))&15;
    if (tab_seg[n]!=0) {
      ptr1=MK_FP(tab_seg[n],0x0459);  /* test MANHOL: use KEYBUF or DKYBUF ? */
      if (((*ptr1)&0x1)==0) {
        ptr=MK_FP(tab_seg[n],0x05d0);  /* debugger display buffer */
        fdbg=1;
      }
    }
  }


  faff=0; fcsr=0; flg=0;
  _fmemcpy(ligne,ptr,96);
  if (fdbg) ligne[30]=0; else ligne[96]=0;
  if (strcmp(ligne,derligne)!=0) {
    strcpy(derligne,ligne);
    faff=1;
    derdeb=0;
  }
  /* cursor position */
  xp=*(ptr-1);
  if ((xp!=derpos)||faff) {
    deb=derdeb;
    if (xp<40)
      deb=0;
    else if (xp<deb)
      deb=xp;
    else if (xp>(deb+78))
      deb=xp-78;
    derpos=xp;
    if (deb!=derdeb) {
      derdeb=deb;
      faff=1;
    }
    fcsr=1;
  }
  else
    deb=derdeb;

  if (faff) {
    if (derdeb<(96-79))
      ligne[deb+79]=0;
    fcsr=1;
  }
  /* cursor status */
  xs=*(ptr-4);
  xs &= 0xc0;
  if (xs!=deretat) {
    fcsr=1;
    deretat=xs;
  }



  /* flags */
  /* display driver at $2e100 */
  ptr=MK_FP71(0x2e100L);   /* ANNAD1 */
  flags[0]= *ptr++;
  flags[1]= *ptr;
  /* display driver at $2e34c */
  ptr=MK_FP71(0x2e34cL);   /* ANNAD3 */
  flags[2]= *ptr++;
  flags[3]= *ptr;
  flags[4]=frun;
  if (memcmp(flags,derflags,5)!=0) {
    memcpy(derflags,flags,5);
    flg=1;
  }

  if (faff) {
    update_display(&ligne[deb], NULL, xp-deb+1, xs);
    fcsr=0;
  }
  if (fcsr) {
    update_display(NULL, NULL, xp-deb+1, xs);
  }
  if (flg) {
    update_display(NULL, flags, xp-deb+1, xs);
  }


}


/* ************************************************************* */
/* timer management                                              */
/* ************************************************************* */



/* ****************************************** */
/* init_timer()                               */
/*                                            */
/* init the timer emulation system            */
/* ****************************************** */
void init_timer(void)
{
  last_timer = clock()*UNIT_TICKS;
}


/* ****************************************** */
/* add_targettime()                           */
/*                                            */
/* add a number of tick to the software RTC   */
/* ****************************************** */
void static add_targettime(long k)
{
  unsigned char far *ptr;
  long n, n0;
  int n1;

  ptr=MK_FP71(0x2f70dL);       /*  NXTIRQ */

  n0 = * ((long far *)(ptr));

  n=n0+(k*16);         /* x16 cause NXTIRQ is not aligned on an even boundary */

  * (long far *)(ptr) = n;

  if (n<n0) {
    n1 = * (int far *)(ptr+4);
    n1++;
    * (int far *)(ptr+4) = n1;
  }

}




/* ****************************************** */
/* decr_timer1()                              */
/*                                            */
/* decrement one timer                        */
/* set timer_to=1 if timer goes from 0 to <0  */
/* ****************************************** */
void static decr_timer1(unsigned char far *ptr)
{
  long n;

  n = *(ptr+2);
  n <<= 8;
  n += *(ptr+1);
  n <<=8;
  n += *ptr;

  if (n==0) timer_to=1;
  n--;

  *ptr = n;
  *(ptr+1) = n >> 8;
  *(ptr+2) = n >> 16;

  return;
}


/* ****************************************** */
/* decr_timers()                              */
/*                                            */
/* decrement the 3 timers                     */
/*  timer_to is set to 1 if timer OVF         */
/* ****************************************** */
void static decr_timers(void)
{
  unsigned char far *ptr;

  ptr = MK_FP71(0x2e1f8L);/* TIMER3 */
  decr_timer1(ptr);
  ptr += 0x80;            /* TIMER2 */
  decr_timer1(ptr);
  ptr += 0x80;            /* TIMER1 */
  decr_timer1(ptr);
}




/* ****************************************** */
/* check_timer()                              */
/*                                            */
/*  check timers for SRQ                      */
/*  returns 1 if one timer as MSB=1           */
/* ****************************************** */
int static check_timer(void)
{
  unsigned char far *ptr;
  int n;

  n=0;
  ptr = MK_FP71(0x2e1f8L);   /* TIMER3 */
  if ((*(ptr+2))&(*(ptr+3))&0x80) n=1;
  ptr += 0x80;               /* TIMER2 */
  if ((*(ptr+2))&(*(ptr+3))&0x80) n=1;
  ptr += 0x80;               /* TIMER1 */
  if ((*(ptr+2))&(*(ptr+3))&0x80) n=1;
  return(n);
}


/* ****************************************** */
/* check_ilmb()                               */
/*                                            */
/*  check one HPIL mailbox                    */
/*  returns 1 if mailbox needs service        */
/* 04/01/2009                                 */
/* ****************************************** */
int static check_ilmb(int n)
{
  unsigned char far *ptr;

  ptr = MK_FP(tab_seg[2],n*8);
  if ((*(ptr+3))&0x90) return(1);     /* fixed (?) 3/07/2010 */
  return(0);
}



/* ************************************************************* */
/* file loading                                                  */
/* ************************************************************* */


/* ****************************************** */
/* loadrom()                                  */
/*                                            */
/*  load the mainframe ROM                    */
/*  install patches needed for Emu71          */
/* ****************************************** */
void loadrom(void)
{
  FILE *fic;
  unsigned char far *ptr;
  int i, n;

  /* init memory */
  fic=fopen(ROM71,"rb");
  if (fic==NULL) {
    fin_ecran();
    fprintf(stderr,"Emu71: file %s not found\n",ROM71);
    exit(0);
  }
  ptr=MK_FP(tab_seg[0],0);
  for (i=0;i<32;i++) {
    n=fread(buffer,1,1024,fic);
    if (n!=1024) break;
    _fmemcpy(ptr,buffer,1024);
    ptr+=1024;
  }
  ptr=MK_FP(tab_seg[1],0);
  for (i=0;i<32;i++) {
    n=fread(buffer,1,1024,fic);
    if (n!=1024) break;
    _fmemcpy(ptr,buffer,1024);
    ptr+=1024;
  }
  fclose(fic);
  _fmemset(ptr,0,8);  /* security */
  if (i!=32) {
    fin_ecran();
    fprintf(stderr,"Emu71: error in file %s\n",ROM71);
    exit(0);
  }

  /* patch for ports 6-7, JFG 26/12/05 */
  /* change 5 by 7 at 09EC0 (in =PDEV, SG&FXQ module) */
  ptr=MK_FP71(0x09ec0L);
  if (*ptr==0x25) {
    *ptr=0x27;
    ptr=MK_FP71(0x0aa82L);  /* checksum 2 at 0AA81-82 */
    n=*ptr; *ptr = n-2;
  }

  /* patch for key debounce, JFG 24/03/06 */
  /* change GONC to GOC at 00D32 (in =DEBNCE, SB&KEY module) */
  ptr=MK_FP71(0x00d32L);
  if (*ptr==0xc5) {
    *ptr=0xc4;
    ptr=MK_FP71(0x0000eL);     /* checksum 1 at 0000D-0E */
    n=*ptr; *ptr = n+1;
  }

  /* patch for CR, JFG 17/04/06:
  write RTNSC instead of RTN at 1ceb7  (in CRPRS?),
  i.e. 28 instead of 18 at 1ceb8 */
  ptr=MK_FP71(0x1ceb8L);
  if (*ptr==0x81) {
    *ptr=0x82;
    ptr=MK_FP71(0x1dba6L);     /* checksum 4 at 1DBA6-A7  */
    n=*ptr; *ptr = n-1;
  }

}


/* ****************************************** */
/* loadmodule()                               */
/*                                            */
/*  load a module ROM                         */
/*  install patches needed for HPIL           */
/* ****************************************** */
void static loadmodule(int seg, int taille, char filename[])
{
  FILE *fic;
  unsigned char far *ptr;
  int i, n;
  char buf[80];

  /* init memory */
  for (i=0;i<taille;i++) {     /* JFG 08/03/03 */
    ptr=MK_FP(seg+i*0x40,0);
    _fmemset(ptr,0,0x400);
  }

  fic=fopen(filename,"rb");
  if (fic==NULL) {
    sprintf(buf,"Emu71: file %s not found\r\n",filename);
    disp_string(buf);
  }
  else {
    for (i=0;i<taille;i++) {
      ptr=MK_FP(seg+i*0x40,0);
      n=fread(buffer,1,0x400,fic);
      if (n!=0x400) break;
      _fmemcpy(ptr,buffer,0x400);
    }
    fclose(fic);

    /* test if HPIL rom */
    if ( (strstr(filename,"HPIL")!=NULL)
      || (strstr(filename,"hpil")!=NULL) ) {
      /* yes, install patches */
      /* replace 15c7 with 154f at 6ac2
                 15e6 with 156e at 6758 and 67f9 */
      ptr=MK_FP(seg,0x3562);
      if ((*ptr)==0x7c) *ptr = 0xf4;
      n = 0xf4-0x7c;
      ptr=MK_FP(seg,0x33ad);
      if ((*ptr)==0x6e) *ptr = 0xe6;
      n += 0xe6 - 0x6e;
      ptr=MK_FP(seg,0x33fd);
      if ((*ptr)==0xe5) { *ptr = 0x65; ptr++; *ptr=0x8e; }  /* was 0x86 */
      n += 0x65 - 0xe5 + 0x8e - 0x86;
      /* 3/07/10: add a patch for RESET HPIL */
      /* replace 15c8 with 154f at 6df9 */
      ptr=MK_FP(seg,0x36fd);
      if ((*ptr)==0xc5) { *ptr = 0x45; ptr++; *ptr=0x7f; }  /* was 0x78 */
      n += 0x45 - 0xc5 + 0x7f - 0x78;
      /* update checksum */
      ptr=MK_FP(seg,0x3ffe);
      n = (*ptr - n);
      *ptr = n;
    }

  }
}






/* ************************************************************* */
/* initialization                                                */
/* ************************************************************* */


/* ****************************************** */
/* init_register()                            */
/*                                            */
/*   init the CPU registers                   */
/* ****************************************** */
void init_register(void)
{
  int i;

  for (i=0;i<2;i++) {
    regA[i]=0;
    regB[i]=0;
    regC[i]=0;
    regD[i]=0;
    regR0[i]=0;
    regR1[i]=0;
    regR2[i]=0;
    regR3[i]=0;
    regR4[i]=0;
  }
  Carry=0;
  Pointer=0;
  hwstatus=0;
  flagdec=0;
  retstkptr=0;
  regD0=0;
  regD1=0;
  regPC=0;
  for (i=0;i<8;i++) retstk[i]=0;
  status=0;
  input=0;
  output=0;
}


/* ****************************************** */
/* init_ram()                                 */
/*                                            */
/*   init the RAM                             */
/* ****************************************** */
void init_ram(void)
{
  unsigned char far *ptr;
  int i;

  /* initialize RAM to zro */
  /* I/O & sys RAM */
  ptr=MK_FP(tab_seg[2],0);
  for (i=0;i<32;i++) {
    _fmemset(ptr,0,1024);
    ptr+=1024;
  }

  /* Emu71 special locations:
  2f000-2f00f: date&time (cf maj_date)
  2f010-2f01f: Emu71 signatures
  */


}



/* ****************************************** */
/* maj_date()                                 */
/*                                            */
/*  set the date & time in Emu71 reserved     */
/*   RAM locations                            */
/*  also set the emulator signature           */
/* ****************************************** */
void static maj_date(void)
{
  struct date d;
  struct time t;
  int i, n;
  long ld, lt;
  long far *pt;

  /* write date&time in 2f000 */
  /* date = year*1000 + day of year
  /* time = number of seconds since midnight */

  getdate(&d);
  n = d.da_day;
  for (i=1;i<d.da_mon;i++) {
    switch (i) {
      case 2:
        if (d.da_year%4) n += 28; else n += 29; /* warning: will fail in 2100 ! */
        break;
      case 4:
      case 6:
      case 9:
      case 11:
        n += 30;
        break;
      default:
        n += 31;
    }
  }
  ld = (d.da_year%100)*1000L + n;

  gettime(&t);
  lt=t.ti_hour;
  lt *=60;
  lt += t.ti_min;
  lt *= 60;
  lt += t.ti_sec;

  pt = MK_FP71(0x2f000L);
  *pt++ = ld;
  *pt = lt;

  /* emulator ID */
  pt = MK_FP71(0x2f010L);
  *pt = 0x42131L;       /* DOS platform, Emu71 ext set, x86, v2.4 */
  /* Platform/emulator/cpu information:
    2f010 Platform ID
      0: HP-71B
      1: Emu71  (DOS)
      2: Emu71w (Windows)
      3: Emu71m (Mobile)
    2f011 Saturn Opcode Level
      0: 1LF2 (HP-71B set)
      1: 1LK7 (HP-28C, PC=(A) and RSI added)
      2: 1LR2 (HP-48S/G)
      3: Emu71 extended set
    2f012 Host CPU type
      0: Saturn
      1: x86 (Intel 16 bits)
      2: i32 (Intel 32 bits)
      3: SH3
      4: ARM
    2f013 Emulator version
    2f014 Emulator revision
  */
}


/* ****************************************** */
/* init_hrddbg()                              */
/*                                            */
/*  init the hardwired debugger signature     */
/*   (for the DEBUG keyword in Ass/forth ROM  */
/* ****************************************** */
void static init_hrddbg(void)
{
  char far *ptr;

  ptr=MK_FP71(0xf0000L);
  _fmemset(ptr,0,4096);
  _fmemcpy(ptr,"X:GBD ",6);
}



/* ****************************************** */
/* loaddevice()                               */
/*                                            */
/*  load and init a HPIL device               */
/* ****************************************** */
void static loaddevice(char *nomdev, char *option)
{
  int tdev;
  int d, n;

  if      (stricmp(nomdev,"DISPLAY")==0) tdev=1;
  else if (stricmp(nomdev,"HDRIVE1")==0) tdev=2;
  else if (stricmp(nomdev,"FDRIVE1")==0) tdev=3;
  else if (stricmp(nomdev,"DOSLINK")==0) tdev=4;
  else if (stricmp(nomdev,"XIL")==0) {
    if ((option[0]=='c')||(option[0]=='C'))
     tdev=10;
    else
     tdev=5;
  }
  else if (stricmp(nomdev,"LPRTER1")==0) tdev=6;
  else if (stricmp(nomdev,"HDRIVE2")==0) tdev=7;
  else if (stricmp(nomdev,"SERIAL1")==0) tdev=8;
  else if (stricmp(nomdev,"SERIAL2")==0) tdev=9;
  /* XIL COMx is tdev=10 */
/*  else if (stricmp(nomdev,"PILPIO1")==0) tdev=11; */
  else tdev=0;

  if (tdev && (nbdev<32)) {
    tabdev[nbdev]=tdev;
    nbdev++;
  }
  switch (tdev) {
    case 1:
      init_ildisplay();
      break;
    case 2:
      init_ilhdisc(option);
      break;
    case 3:
      if ((option[0]=='b')||(option[0]=='B')) d=1; else d=0;
      init_ilfdisc(d);
      break;
    case 5:
#ifdef XIL
      d=0;
      if (option[0]!=0) sscanf(option,"%x",&d);
      if (d==0) d=0x1700;
      init_ilext(d,1);
      f_xil=1;
#endif
      break;
    case 6:
      init_ilcentro();
      break;
    case 7:
      init_ilhdisc2(option);
      break;
    case 8:
      d=0;
      n=0xe3;
      if (option[0]!=0) {
        if (option[1]==0) {
          sscanf(option,"%d",&d);
          d--;
        }
        else if (option[1]==',') {
          sscanf(option,"%d,%x",&d,&n);
          d--;
        }
        else {
          sscanf(option,"%x",&n);
        }
      }
      init_ilserial1(d,n);
      break;
    case 9:
      d=1;
      n=0xe3;
      if (option[0]!=0) {
        if (option[1]==0) {
          sscanf(option,"%d",&d);
          d--;
        }
        else if (option[1]==',') {
          sscanf(option,"%d,%x",&d,&n);
          d--;
        }
        else {
          sscanf(option,"%x",&n);
        }
      }
      init_ilserial2(d,n);
      break;
    case 10:  /* XIL2 */
      if ((option[1]!=0)&&(option[2]!=0)&&(option[3]!=0)) {
        sscanf(&option[3],"%d",&d);
        init_ilext2(d,1);
      }
      break;
/*    case 11:  /* PIO1 
      init_ilpio1();
      break;
*/
  }

}






/* ****************************************** */
/* init_emulate()                             */
/*                                            */
/* read the file EMU71.INI                    */
/* initialize structures device and tab_seg   */
/* ****************************************** */
void static init_emulate(int dbg)
{
  unsigned char far *ptrmem;
  int i, n, nbre, ndeb, seg;
  FILE *fic;
  char ligne[82];
  int nuport, taille, taille0;
  char nomfic[40];
  char type[8];
  char nomdev[40];
  char option[40];
  char f;

  for (i=0;i<16;i++)
    tab_seg[i]=0;
  for (i=0;i<16;i++)
    device71[i].type=0;

  ndeb=0;

  device71[0].nuport = 0;
  device71[0].type = 'A';  /* RAM */
  device71[0].taille= RAMSIZE;
  strcpy(device71[0].nomfic , "RAM71.BIN");
/* id RAM = $8000x
 x: c   4Ko
    b   8Ko
    a  16Ko
    9  32Ko
    8  64Ko (max.)
*/
  device71[0].id= 0x80009L;
  if (RAMSIZE==64) device71[0].id -= 1;       /* JFG 01/03/03 */
  if (RAMSIZE==128) device71[0].id -= 2;       /* JFG 08/03/03 */
/*  device71[0].seg=tab_seg[3]; */

  ndeb=1;


  /* read file .ini, and fill struct device */
  f=1;
  fic=fopen("emu71.ini","rt");
  if (fic==NULL) {
    fin_ecran();
    printf("Emu71: file emu71.ini not found !\n");
    exit(0);
  }
  else {
    nbre=ndeb;
    taille0=0;
    while (fgets(ligne,80,fic)) {
      if ( (ligne[0]!=';') && (ligne[0]!=' ') && (ligne[0]!='\n') ) {
        if (ligne[0]=='[') {
          if (strnicmp(&ligne[1],"MOD",3)==0)
            f=1;
          else if (strnicmp(&ligne[1],"DEV",3)==0)
            f=2;
          else if (strnicmp(&ligne[1],"END",3)==0)
            f=3;
        }
        else {
          switch (f) {
          case 1:  /* modules */
            n=sscanf(ligne,"%d %s %d %s",&nuport,type,&taille,nomfic);
            if ((n==4)&&(nbre<=14)) {
              if ((type[1]=='R')&&(nuport==0))
                strcpy(ROM71,nomfic);
              else {
                if ((type[1]=='A')&&(nuport==0)) {
                  if ((taille0==0)&&(nbre==1)) nbre=0;
                  taille0 +=taille;
                }
                device71[nbre].nuport = nuport;
                device71[nbre].type = type[1];
                device71[nbre].taille=taille;
                strcpy(device71[nbre].nomfic , nomfic);
                if (type[1]=='O') device71[nbre].id= 0x80109L;
                else if (type[1]=='A') device71[nbre].id= 0x80009L;
                else if (type[1]=='E') device71[nbre].id= 0x80209L; /* JFG 30/04/12 */
                if (taille==64) device71[nbre].id -= 1;       /* JFG 01/03/03 */
                if (taille==128) device71[nbre].id -= 2;       /* JFG 08/03/03 */
                nbre++;
              }
            }
            break;
          case 2:  /* devices */
            option[0]=0;
            n=sscanf(ligne,"%s %s",nomdev,option);
            if (n>=1) {
              loaddevice(nomdev, option);
            }
            break;
          }
        }
      }
    }

    fclose(fic);

    /* allocate HPIL mailboxes ( at least 2 possibles, as nbre=14 max here) */
    /* HP-IL */
    device71[nbre].nuport = 0;
    device71[nbre].type = 'X';  /* I/O */
    device71[nbre].taille= 1;
    strcpy(device71[nbre].nomfic , "HPIL");
    /* id I/O = $80F0F
      1 word (16 nibs), memory mapped I/O, HPIL mailbox
    */
    device71[nbre].id= 0x80F0FL;
    nbre++;

  }


  /* memory allocation */

  /* --- internal ROM --- */

  /* try to allocate internal ROM in EMS to save memory */
  seg=GetEmsSegment();
  if (seg!=0) {
    tab_seg[0]=seg;
    tab_seg[1]=tab_seg[0]+0x0800;
    if (dbg) fprintf(stderr,"Emu71: Using EMS\n");
  }
  else {
    if (dbg) fprintf(stderr,"Emu71: No EMS\n");
    ptrmem=farmalloc(0x10020L);

    if (ptrmem==NULL) {
      fin_ecran();
      fprintf(stderr,"Emu71: No enough memory\n");
      exit(0);
    }
    /* security */
    if (FP_OFF(ptrmem)>15) {
      fin_ecran();
      fprintf(stderr,"Emu71: invalid memory pointer\n");
      exit(0);
    }
    /* ajust for beginning of page */
    tab_seg[0]=FP_SEG(ptrmem)+1;
    tab_seg[1]=tab_seg[0]+0x0800;
  }

  loadrom();

  /* --- I/O, system RAM, RAM --- */

  if (taille0==0) taille0=RAMSIZE;
  ptrmem=farmalloc(0x9020L+0x400L*taille0);

  if (ptrmem==NULL) {
    fin_ecran();
    fprintf(stderr,"Emu71: No enough memory\n");
    exit(0);
  }
  /* security */
  if (FP_OFF(ptrmem)>15) {
    fin_ecran();
    fprintf(stderr,"Emu71: invalid memory pointer\n");
    exit(0);
  }
  /* ajust for beginning of page */
  seg=FP_SEG(ptrmem)+1;
  tab_seg[15]=seg;   /* use same segment than I/O */
  seg += 0x100;
  tab_seg[2]=seg;
  seg += 0x800;

  /* HRDDBG */
  init_hrddbg();


  init_ram();

  init_hpil();

  /* allocate memory segments, and read image files */
  nbre=0;
  while (device71[nbre].type) {
    if (device71[nbre].type!='X') {

      taille=32;   /* RAM */
      if (device71[nbre].taille==128) taille=128;  /* JFG 08/03/03 */
      if (device71[nbre].taille==64) taille=64;  /* JFG 01/03/03 */
      if ((device71[nbre].nuport==0)&&device71[nbre].type=='A') {
        device71[nbre].seg=seg;
        seg += device71[nbre].taille * 0x40;
      }
      else {
        ptrmem=farmalloc(1024L*taille+32);
        if (ptrmem==NULL) {
          fin_ecran();
          fprintf(stderr,"Emu71: No enough memory for %s\n",
                         device71[nbre].nomfic);
          exit(0);
        }
        /* security */
        if (FP_OFF(ptrmem)>15) {
          fin_ecran();
          fprintf(stderr,"Emu71: invalid memory pointer\n");
          exit(0);
        }
        /* ajust for beginning of page */
        device71[nbre].seg=FP_SEG(ptrmem)+1;
      }

      loadmodule(device71[nbre].seg, device71[nbre].taille,
                 device71[nbre].nomfic);

      /* hardconfig ROM at E0000 */
      if (device71[nbre].type == 'R')
        tab_seg[14] = device71[nbre].seg;

    }

    else
      device71[nbre].seg=tab_seg[2];

    nbre++;

  }


}





/* ************************************************************* */
/* save/restore of system state                                  */
/* ************************************************************* */


/* ****************************************** */
/* save_state()                               */
/*                                            */
/* save state of processor and RAM I/O        */
/*   in sys71.dat,                            */
/* RAM in ram71.bin, and IRAM in  files       */
/* ****************************************** */
void save_state(void)
{
  FILE *fic;
  char far *ptr;
  struct portdevice *deviceptr;
  int i, j;
  long hrstk[8];  /* "hardware" standard 8 level stack */
  long xrstk[56]; /* 56 level stack extension */

  /* First 8 levels (for compatibility with standard stack ) */
  j=retstkptr;
  for (i=0;i<8;i++) {
    hrstk[i]=retstk[j];
    j++; if (j==64) j=0;
  }
  /* 56 level extension */
  for (i=0;i<56;i++) {
    xrstk[i]=retstk[j];
    j++; if (j==64) j=0;
  }

  /* save extended Emu71 Saturn state: */
  /* R5-R7 are in 2F200-2F22F */
  ptr=MK_FP71(0x2f200L);
  _fmemcpy(ptr,regR5,8);
  _fmemcpy(ptr+8,regR6,8);
  _fmemcpy(ptr+16,regR7,8);
  /* the 56 extra stack levels are in 2F240-2F3FF */
  _fmemcpy(ptr+32,xrstk,4*56);

  /* sys71.dat */
  fic=fopen("sys71.dat","wb");
  if (fic==NULL) {
    fin_ecran();
    printf("Emu71: error writing sys71.dat\n");
    exit(0);
  }
  /*  processor state */
  fwrite(regA,sizeof(regsat),1,fic);
  fwrite(regB,sizeof(regsat),1,fic);
  fwrite(regC,sizeof(regsat),1,fic);
  fwrite(regD,sizeof(regsat),1,fic);
  fwrite(regR0,sizeof(regsat),1,fic);
  fwrite(regR1,sizeof(regsat),1,fic);
  fwrite(regR2,sizeof(regsat),1,fic);
  fwrite(regR3,sizeof(regsat),1,fic);
  fwrite(regR4,sizeof(regsat),1,fic);
  fwrite(&regD0,sizeof(long),1,fic);
  fwrite(&regD1,sizeof(long),1,fic);
  fwrite(&regPC,sizeof(long),1,fic);
/*  fwrite(retstk,sizeof(long),8,fic);  changed JFG 23/07/07 */
  fwrite(hrstk,sizeof(long),8,fic);  /* first 8 level hardware stack */
  fwrite(&status,sizeof(short),1,fic);
  fwrite(&input,sizeof(short),1,fic);
  fwrite(&output,sizeof(short),1,fic);
  fwrite(&Carry,sizeof(char),1,fic);
  fwrite(&Pointer,sizeof(char),1,fic);
  fwrite(&hwstatus,sizeof(char),1,fic);
  fwrite(&flagdec,sizeof(char),1,fic);
/*  fwrite(&retstkptr,sizeof(char),1,fic);    changed JFG 23/07/07 */
  i=0;  fwrite(&i,sizeof(char),1,fic);  /* write retstkptr=0 */

  /* RAM I/O : 2e000 - 2ffff*/
  ptr=MK_FP71(0x2e000L);
/*  fwrite(ptr,1,0x1000,fic); */
  for (i=0;i<4;i++) {
    _fmemcpy(buffer,ptr,0x400);
    fwrite(buffer,1,0x400,fic);
    ptr+=0x400;
  }
  fclose(fic);


  /* RAM and EEPROM (JFG 30/04/12) */
  deviceptr = &device71[0];
  for (i=0;i<16;i++) {
    if (deviceptr->type==0) break;  /* end of devices */
    if ( (deviceptr->type=='A') || (deviceptr->type=='E') ) {
      fic=fopen(deviceptr->nomfic,"wb");
      if (fic==NULL) {
        fin_ecran();
        printf("Emu71: error writing %s\n",deviceptr->nomfic);
        exit(0);
      }

      /* JFG 08/03/03 */
      for (j=0;j<deviceptr->taille;j++) {
        ptr=MK_FP(deviceptr->seg+j*0x40,0);
        _fmemcpy(buffer,ptr,0x400);
        fwrite(buffer,1,0x400,fic);
      }


      fclose(fic);
    }
    deviceptr++;
  }

}


/* ****************************************** */
/* load_state()                               */
/*                                            */
/* restore processor state and RAM I/O        */
/* from sys71.dat                             */
/* ****************************************** */
void load_state(void)
{
  FILE *fic;
  char far *ptr;
  int i, j;
  char buf[80];
  long hrstk[8];
  long xrstk[56];

  /* sys71.dat */
  fic=fopen("sys71.dat","rb");
  if (fic==NULL) {
    sprintf(buf,"Emu71: error reading sys71.dat\r\n");
    disp_string(buf);
    return;
  }
  /*  processor state */
  fread(regA,sizeof(regsat),1,fic);
  fread(regB,sizeof(regsat),1,fic);
  fread(regC,sizeof(regsat),1,fic);
  fread(regD,sizeof(regsat),1,fic);
  fread(regR0,sizeof(regsat),1,fic);
  fread(regR1,sizeof(regsat),1,fic);
  fread(regR2,sizeof(regsat),1,fic);
  fread(regR3,sizeof(regsat),1,fic);
  fread(regR4,sizeof(regsat),1,fic);
  fread(&regD0,sizeof(long),1,fic);
  fread(&regD1,sizeof(long),1,fic);
  fread(&regPC,sizeof(long),1,fic);
  fread(hrstk,sizeof(long),8,fic);  /* changed JFG 23/07/07 */
  fread(&status,sizeof(short),1,fic);
  fread(&input,sizeof(short),1,fic);
  fread(&output,sizeof(short),1,fic);
  fread(&Carry,sizeof(char),1,fic);
  fread(&Pointer,sizeof(char),1,fic);
  fread(&hwstatus,sizeof(char),1,fic);
  fread(&flagdec,sizeof(char),1,fic);
  fread(&retstkptr,sizeof(char),1,fic);

  /* RAM I/O : 2e000 - 2ffff*/
  ptr=MK_FP71(0x2e000L);
/*   fread(ptr,1,0x1000,fic); */
  for (i=0;i<4;i++) {
    fread(buffer,1,0x0400,fic);
    _fmemcpy(ptr,buffer,0x400);
    ptr += 0x400;
  }
  fclose(fic);
  /* restore extended Emu71 Saturn state: */
  /* R5-R7 are in 2F200-2F22F */
  ptr=MK_FP71(0x2f200L);
  _fmemcpy(regR5,ptr,8);
  _fmemcpy(regR6,ptr+8,8);
  _fmemcpy(regR7,ptr+16,8);
  /* the 56 extra stack levels are in 2F240-2F3FF */
  _fmemcpy(xrstk,ptr+32,4*56);

  /* restaure full 64 level stack */
  j=retstkptr;
  for (i=0;i<8;i++) {
    retstk[i]=hrstk[j];
    j++; if (j==8) j=0;
  }
  /* 56 level extension */
  j=0;
  for (i=8;i<64;i++) {
    retstk[i]=xrstk[j];
    j++;
  }
  retstkptr=0;  /* init stack ptr to first item */

}



/* ****************************************** */
/* checkrom()                                 */
/*                                            */
/* check the checksum of ROMs                 */
/* ****************************************** */
void checkrom(void)
{
  int i, j, k, n, m;
  unsigned char far *ptr;
  unsigned char *ptr1;
  char nom[40];
  struct portdevice *deviceptr;

  /* test internal ROM */
  ptr=MK_FP(tab_seg[0],0);
  printf("00000 ROM TEST");
  for (i=0;i<4;i++) {
    n=0;
    printf(" %d",i+1);
    for (j=0;j<16384;j++) {
      n += *ptr++;
      if (n>255) {
        n++;
        n &= 255;
      }
    }
    n--;
    if (n==0) printf("G"); else printf("B(%x)",n);
  }
  printf("\n");

  /* test module ROM or EEPROM */
  deviceptr = &device71[0];
  for (i=0;i<16;i++) {
    if (deviceptr->type==0) break;  /* end of devices */
    m=deviceptr->type;
    if ((m=='O')||(m=='E')) {  /* ROM or EEPROM */
      strcpy(nom,deviceptr->nomfic);
      ptr1=strchr(nom,'.');
      if (ptr1) *ptr1=0;
      ptr1=strchr(nom,'\\');
      if (ptr1) ptr1++; else ptr1=nom;
      strupr(ptr1);
      printf("%05lX %-8s",deviceptr->adresse,ptr1);
      ptr=MK_FP(deviceptr->seg,0);
      m=*ptr;
      n=0;
      for (k=0;k<(deviceptr->taille/16);k++) {
        printf(" %d",k+1);
        for (j=0;j<16384;j++) {
          n += *ptr;
          ptr++;
          if (n>255) {
            n++;
            n &= 255;
          }
        }
        if (m!=0x3B) {
          n--;
          if (n==0) printf("G"); else printf("B(%x)",n);
          n=0;
        }
      }
      if (m) {
        n--;
        if (n==0) printf(" G"); else printf(" B(%x)",n);
      }
      printf("\n");
    }
    deviceptr++;
  }

  /* test Forth ROM */
  if (tab_seg[14]) {
    ptr=MK_FP(tab_seg[14],0);
    if ((*ptr)==0x46) {
      printf("E0000 HRDFORTH");
      for (i=0;i<2;i++) {
        n=0;
        printf(" %d",i+1);
        for (j=0;j<16384;j++) {
          n += *ptr++;
          if (n>255) {
            n++;
            n &= 255;
          }
        }
        n--;
        if (n==0) printf("G"); else printf("B");
      }
      printf("\n");
    }
  }

}





/* ************************************************************* */
/* Emulation                                                     */
/* ************************************************************* */



/* ****************************************** */
/* execute_special()                          */
/*                                            */
/* execution of the special Saturn instructions: */
/*   configuration, SREQ, shutdown            */
/* and HPIL special opcodes                   */
/* ****************************************** */
int execute_special(int code)
{
   int i, n, ret, ok;
   struct portdevice *deviceptr;
   unsigned char far *ptr;
   char buf[80];

   ret=0;
   switch (code) {

     case RESET:
       if (f_trace&1) {
         sprintf(buf,"Trace: RESET\r\n");
         disp_string(buf);
       }
       n=0;
       for (i=0;i<16;i++) {
         device71[i].configure=0;
         if (device71[i].type == 'R') n=1; /* test if hard forth present */
       }
       for (i=3;i<14;i++)
         tab_seg[i]=0;
       if (n==0) tab_seg[14]=0;  /* reserved ROM forth hard */
       break;

     case ID:
       regC[0] &= 0xfff00000L;
       deviceptr = &device71[0];
       for (i=0;i<16;i++) {
         if (deviceptr->type==0) break;  /* end of devices */
         if (deviceptr->nuport==0)
           ok = 1;
         else
           ok = (1<<(deviceptr->nuport-1)) & output;
         ok = ok && (deviceptr->type!='R');  /* <> hardconfig */
         ok = ok && (deviceptr->configure==0);
         if (ok) {
           regC[0] |= deviceptr->id;
           break;
         }
         deviceptr++;
       }
       if (f_trace&1) {
         sprintf(buf,"Trace: ID  output=%03x C(A)=%05lx\r\n",
                      output,regC[0]&0xfffffL);
         disp_string(buf);
       }
       break;

     case CONFIG:
       deviceptr = &device71[0];
       for (i=0;i<16;i++) {
         if (deviceptr->type==0) break;  /* end of devices */
         if (deviceptr->nuport==0)
           ok = 1;
         else
           ok = (1<<(deviceptr->nuport-1)) & output;
         ok = ok && (deviceptr->configure==0);
         ok = ok && (deviceptr->type!='R');  /* <> hardconfig */
         if (ok) {
           deviceptr->adresse = regC[0] & 0xfffffL;
           deviceptr->configure = 1;
           n=deviceptr->adresse >> 16;
           if ( (n>2) && (n<15) ) {
             tab_seg[n]=deviceptr->seg;
             if (deviceptr->taille>=64)  /* JFG 01/03/03 */
               tab_seg[n+1]=deviceptr->seg+0x800;
             if (deviceptr->taille==128) { /* JFG 08/03/03 */
               tab_seg[n+2]=deviceptr->seg+0x1000;
               tab_seg[n+3]=deviceptr->seg+0x1800;
             }
           }
           break;
         }
         deviceptr++;
       }
       if (f_trace&1) {
         sprintf(buf,"Trace: CONFIG  output=%03x C(A)=%05lx\r\n",
                           output,regC[0]&0xfffffL);
         disp_string(buf);
       }
       break;

     case UNCNFG:
       deviceptr = &device71[0];
       for (i=0;i<16;i++) {
         if (deviceptr->type==0) break;  /* end of devices */
         if ( (deviceptr->configure) && (deviceptr->type!='R') ) {
           if (deviceptr->adresse == (regC[0] & 0xfffffL)) {
             n=deviceptr->adresse >> 16;
             tab_seg[n]=0;
             if (deviceptr->taille>=64)  /* JFG 01/03/03 */
               tab_seg[n+1]=0;
            if (deviceptr->taille==128) { /* JFG 08/03/03 */
               tab_seg[n+2]=0;
               tab_seg[n+3]=0;
             }
             deviceptr->configure = 0;
           }
         }
         deviceptr++;
       }
       if (f_trace&1) {
         sprintf(buf,"Trace: UNCNFG  C(A)=%05lx\r\n",regC[0]&0xfffffL);
         disp_string(buf);
       }
       break;

     case SREQ:
       /* no more called - done in sat.obj */
       regC[0] &= 0xfffffff0L;
       n=check_timer();
       if (n) {
         regC[0] |= 1;
         hwstatus |= 4;  /* SR */
       }
       /* HPIL SRQ is not handled, seems to be used only for remote kbd, which is useless in Emu71 */
       /* 0750c, b311C */
       break;

     case SHUTDN:
       ret=1;
       break;

     case RTNSXM:
       /* now called only if breakpoint or debugger */
       regPC -= 2;
       if (regPC==0xf0010L) {
         /* call to debugger ! */
         regPC=retstk[retstkptr];
         retstk[retstkptr]=0;
         retstkptr = (retstkptr+1) & 63;
         ret=-4;
       }
       else if (regPC==lastaddr) {
         /* breakpoint ! */
         ret= -4;
       }
       else {
         /* normal rtnsxm (no more called - done in sat.obj)*/
         regPC=retstk[retstkptr];
         retstk[retstkptr]=0;
         retstkptr = (retstkptr+1) & 63;
         hwstatus |= 1; /* XM */
       }
       break;

     case WRIL1:
       n=0;
       goto WRILx;
     case WRIL2:
       n=0x10;
       goto WRILx;
     case WRIL3:
       n=0x20;
       goto WRILx;
     case WRIL:      /* called after DAT0=C 8 to a HPIL mailbox*/
       n=regD0&0x30;  /* mailbox number */
WRILx:
       ptr=MK_FP(tab_seg[2],n/2);
       if (f_trace&4) {
         sprintf(buf,"WRIL: %02x %02x %02x %02x\r\n",*ptr,*(ptr+1),*(ptr+2),*(ptr+3));
         disp_string(buf);
       }
       _fmemcpy(buffer,ptr,8);
       ilmailbox(n>>4,buffer);
       _fmemcpy(ptr,buffer,8);
       break;

     case RDIL:      /* called after C=DAT0 7 from a HPIL mailbox */
       n=regD0&0x30;  /* mailbox number */
       ptr=MK_FP(tab_seg[2],n/2);
       if (f_trace&4) {
         sprintf(buf,"RDIL: %02x %02x %02x %02x\r\n",*(ptr+4),*(ptr+5),*(ptr+6),*(ptr+7));
         disp_string(buf);
       }
       *(ptr+4) &= 0xfe;  /* reset flag data */
       _fmemcpy(buffer,ptr,8);
       ilmailbox(n>>4,buffer);
       _fmemcpy(ptr,buffer,8);
       break;

     default:
       ret=-5;
    }

   return(ret);
}







/* ****************************************** */
/* simulator()                                */
/*                                            */
/* simulator high level main loop             */
/* returns 1 if ended by deep sleep,          */
/*   or 2 if call to debugger                 */
/* ****************************************** */
int simulator(void)
{
  int n, k;
  long t, t0;
  int nb_ticks;
  int ret;

  t0=clock()*UNIT_TICKS;
  nb_ticks=0;
  ret=0;
  do {

    do {
      n=executeSaturn(speed_emu,0);
      if (nb_ticks>0) {
        nb_ticks--;
        decr_timers();
      }
    } while ((nb_ticks>0)&&(n==0));
    if (n) {
      n=execute_special(n);
      if (n==1) {
        /* shutdown */
        if (regPC==0x005E2) {
          /* deep sleep */
          ret=1;
          break;
        }
        traite_display(0);
        /* wait for key or timeout */
        n=0;
        if (!flowpower) {
          do {
            k=traite_touche(0);
            if (k<0) ret=2;  /* call to debug */
            clock();
            if (tsttimer()) {
              nb_ticks+=28;
              while (nb_ticks>0) {
                nb_ticks--;
                decr_timers();
                n=timer_to;
                timer_to=0;
                if (n) break;
              }
            }
            send_buffer();
#ifdef XIL
	    if (f_xil) {
              if (hpil_ctrl()==0) traite_extil();
            }
            if (hpil_ctrl()==0) traite_extil2();   /* added JFG 14/04/08 */

#endif
          } while ((k==0)&&(n==0));
        }
        else {
          t=clock()*UNIT_TICKS;
          do {
            k=traite_touche(1);
          } while (k==0);
          if (k<0) ret=2;  /* call to debug */
          tsttimer();
          t0=clock()*UNIT_TICKS;
          t=t0-t;  /* elapsed time in ms */
          t=t*64/125; /* ticks to adjust */
          add_targettime(t);
        }
        traite_display(1);
      }
      else if (n<0)
        ret=2;   /* break point or illegal */
    }


    if (tsttimer()) {
      send_buffer();
      nb_ticks+=28;  /* about 55 ms / (1/512 s) */
      traite_display(1);
      k=traite_touche(0);
      if (k<0) ret=2;
      /* check il mailboxes */
      if (check_ilmb(0)) execute_special(WRIL1);
/*      if (check_ilmb(1)) execute_special(WRIL2); */
/*      if (check_ilmb(2)) execute_special(WRIL3); */
    }

  } while (ret==0);

  printf("\n");
  restore_cursor();

  return(ret);
}









/* ****************************************** */
/* usage()                                    */
/*                                            */
/* ****************************************** */
void static usage(void)
{
  printf("Emu71: emulates the HP-71B machine\n");
  printf("Syntax:  EMU71 [options]\n");
  printf("Options:\n");
  printf("         /d : call debugger at startup\n");
  printf("         /m : HP-IL frame monitor\n");
  printf("         /t : trace system instructions\n");
  printf("         /s : slow host (i.e. 286 or HP95LX)\n");
/*  printf("         /f : fast host (Pentium class)\n"); */
  printf("V2.45 (C) 1996,2016 J-F Garnier   http://www.jeffcalc.hp41.eu/emu71\n\n");
  // added for open source release:
  printf("This program is free software; you can redistribute it and/or\n");
  printf("modify it under the terms of the GNU General Public License\n");
  printf("as published by the Free Software Foundation; either version 2\n");
  printf("of the License, or (at your option) any later version.\n\n");
  printf("This program is distributed in the hope that it will be useful,\n");
  printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
  printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
  printf("GNU General Public License for more details.\n\n");
  printf("You should have received a copy of the GNU General Public License\n");
  printf("along with this program; if not, write to the Free Software\n");
  printf("Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n");

  exit(0);
}



/* ****************************************** */
/* main()                                     */
/*                                            */
/* Emu71 main entry                           */
/*   read and set options                     */
/* ****************************************** */

void main(int argc, char *argv[])
{
  int i, n, ret;
  char *arg;
  int dbg;

  _directvideo=0;    /* BIOS call */

  f_trace=0;
  fmonit=0;
  speed_emu=4*256;
  flowpower=0;
  dbg=0;

  /* read options */
  for (i=1;i<argc;i++) {
    arg=argv[i];
    if (arg[0]=='/') {
      switch(arg[1]) {
        case 'd':
          init_label("SEP71");
          dbg=1;
          break;
        case 'l':
          flowpower=1;
          break;
        case 'm':
          n=atoi(&arg[2]);
          if (n==0) n=1;
          fmonit=n;
          init_log();
          break;
        case 't':
          n=atoi(&arg[2]);
          if (n>0) f_trace=n; else f_trace=1;
          break;
        case 's':
          n=atoi(&arg[2]);
          if (n==0) speed_emu=255;
          if (n==1) speed_emu=192;
          if (n>1)  speed_emu=256/n;
          _directvideo=1;
          break;
        case 'f':
          n=atoi(&arg[2]);
          if (n==0) n=4;
          if (n>=1) speed_emu=256*n;
          break;
        case '?':
          usage();
          break;
        default:
          fprintf(stderr,"Emu71: unknown option '%c'\n",arg[1]);
          exit(0);
      }
    }
  }

  prepa_ecran(0);


  init_emulate(dbg);

  init_timer();

  load_state();
  maj_date();
  push_key(43);   /* ON */
  if (dbg)
    ret=commande();
  else {
    ret=simulator(); /* ret=1 if deep sleep , 2 if break */
    if (ret!=1)
      ret=commande(); /* ret=1 if deep sleep, -1 if exit */
  }
  if (ret==1) save_state();

  init_ilext2(0, -1);  /* added JFG 21/06/09 for PILBox 1.2 */

  EndEms();
  fin_ecran();
}

