/* ****************************************** */
/* Emu71: HP-71B software emulator for PC     */
/*                                            */
/* userpc.c   user interface for emu71        */
/* system dependant (PC version)              */
/*                                            */
/* Version 2.45, J-F Garnier 1996-2016        */
/* ****************************************** */




#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <string.h>

#include "version.h"

static int adsp=0;          /* active display: 1: LCD, 2:screen */
/* DISPLAY */
static int curx, cury;      /* cursor position */
static int nlines=24;       /* number of lines */
static int ncols=80;        /* number of cols */
static char fready;         /* display ready */
static char fesc=0;         /* ESC sequence on going */
static char fcur=0;         /* cursor active */
static char fctyp=1;        /* cursor type */
/* LCD */
static int lcpos=1;        /* cursor position */
static char lctyp=1;       /* cursor type */
/* keyboard */
#ifdef KBUF
static char keybuf[2000];
static int nbkey=0;
#endif


void update_cursor(void);


#if __TURBOC__ < 0x200
#define _NOCURSOR 0x2000
#define _NORMALCURSOR 0x607
#define _SOLIDCURSOR 0x007
void _setcursortype(int n)
{
  _AX=0x0100;
  _CX=n;
  geninterrupt(0x10);
}
#endif

void restore_cursor(void)
{
  _setcursortype(_NORMALCURSOR);
}


void LCDwindow(void);



/* ------------------ timer -------------------------------- */

int tsttimer(void)
{
   int far *ptr;
   int n, ret;
   static int last=0;

   ptr=MK_FP(0x40,0x6c);
   n=*ptr;
   if (n==last) ret=0; else ret=1;
   last=n;
   return(ret);
}



/* ------------------ keyboard -------------------------------- */

int tstkey(void)
{
   int far *ptr;
   int n;

   ptr=MK_FP(0x40,0x1a);
   n=*ptr++;
   return(n!=*ptr);
}



int readkey(int fwait)
{
  unsigned char c, k, f;
  static fs=0;

  unsigned char tabcode[128]= {
    /*  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15 */
    0,  0,  0,  0,  0, 85, 89,  0,103,  0,150,  0,  0, 38,  0,  0,
    /* 16  17  18  19  20  21  22  23  24  25  26  27  28  29  30  31 */
    0,  0, 46,  0,102,  0,  0,  0,  0,  0,  0, 43,  0,  0,  0,  0,
    /*      !   "   #   $   %   &   '   (   )   *   +   ,   -   .   / */
       49,151,152,153,137,138,139,123, 36, 37, 28, 56, 55, 42, 54, 14,
    /*  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ? */
       53, 39, 40, 41, 25, 26, 27, 11, 12, 13,140,136,166, 24,167,168,

    /*  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O */
      154,127,145,143,129,115,130,131,132,120,133,134,135,147,146,121,
    /*  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _ */
      122,113,116,128,117,119,144,114,142,118,141,148,  0,149,126,  0,
    /*  `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o */
        0, 15, 33, 31, 17,  3, 18, 19, 20,  8, 21, 22, 23, 35, 34,  9,
    /*  p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~     */
       10,  1,  4, 16,  5,  7, 32,  2, 30,  6, 29,124,  0,125,  0,107
  } ;

#ifdef KBUF
  if (!fwait)
    if (!kbhit()&&(nbkey==0)) return(0);
  while (kbhit()) {
    c=getch();
    if (nbkey<2000) {
      keybuf[nbkey++]=c;
    }
  }
  c=keybuf[0];
  nbkey--;
  for (i=0;i<nbkey;i++)
    keybuf[i]=keybuf[i+1];
#else
  if (!fwait)
    if (!kbhit()) return(0);
/*    if (!tstkey()) return(0); */
  c=getch();
#endif

  if (c==26) {
    return(-1);
  }

  f=0;
  k=0;
  if (c!=0) {
    if (c<128) {
      k=tabcode[c];
      if (k==103) {
        f=1;
#ifdef KBUF
        nbkey=0;
#endif
      }
    }
  }
  else {
#ifdef KBUF
    c=keybuf[0];
    nbkey=0;
#else
    c=getch();
#endif

    switch(c) {
      case 59: if (fs==1) fs=4; else fs=5;  break;     /* F1 */
      case 60: if (fs==2) fs=4; else fs=6;  break;     /* F2 */
      case 61: k=46;  break;     /* F3 */
      case 62: k=109; break;     /* F4 */
      case 63: k=85;  break;     /* F5 */
      case 64: k=86;  break;     /* F6 */
      case 65: k=87;  break;     /* F7 */
      case 66: k=93;  break;     /* F8 */
      case 67: k=71;  break;     /* F9 */
      case 68: k=72;  break;     /* F10 */
      case 84: k=99;  break;     /* sF1 */
      case 85: k=112; break;     /* sF2 */
      case 86: k=102; break;     /* sF3 */
      case 87: k=165; break;     /* sF4 */
      case 88: k=89;  break;     /* sF5 */
      case 89: k=90;  break;     /* sF6 */
      case 90: k=91;  break;     /* sF7 */
      case 91: k=88;  break;     /* sF8 */
      case 92: k=75;  break;     /* sF9 */
      case 93: k=76;  break;     /* sF10 */
      case 133:k=73;  break;     /* F11 */
      case 134:k=74;  break;     /* F12 */
      case 135:k=77;  break;     /* sF11 */
      case 136:k=92;  break;     /* sF12 */
      case 75: k=47;  break;     /* <- */
      case 77: k=48;  break;     /* -> */
      case 72: k=50;  break;     /* ^  */
      case 80: k=51;  break;     /* v  */
      case 71: k=159; break;     /* << */
      case 79: k=160; break;     /* >> */
      case 73: k=162; break;     /* ^^ */
      case 81: k=163; break;     /* vv */
      case 83: k=104; break;     /* -CHAR */
      case 82: k=105; break;     /* I/R   */
      case 145:
      case 147:k=107; break;     /* -LINE */
      case 149:
      case 132: do_recover();
        k=200;    /* pseudo key code, changed 22/04/06 */
        break;     /* ctrl-/ or ctrl-^^ i.e. ON-/ */
      default: k=0;
    }
    f=1;
  }
  if (k) {
    /* update car code */
    if (k<57) {
      if (fs&1) k +=56;
      if (fs&2) k +=112;
    }
    fs=4;  /* reset shift */
  }
  if ((fs&4)||k) {
    /* update display */
    if (adsp!=1) {
      LCDwindow();
      adsp=1;
    }
    if (ncols>40)
      gotoxy(14,2);
    else
      gotoxy(2,2);
    fs &=3;
    switch (fs) {
      case 0:  cprintf("    "); break;
      case 1:  cprintf("f   "); break;
      case 2:  cprintf("   g"); break;
    }
    update_cursor();
  }

  /* if control keys (cursor, etc) :   */
  if (f) {
    /* clear buffer for controlled autorepeat of ctrl keys */
    while (kbhit()) getch();
    if (keybufst()!=0) k=0; /* clear key is HP71 buffer not empty */
  }
  else {
/*    while (kbhit()) getch(); */
  }


  return(k);
}


int iskeydown(void)
{
 return(bioskey(2)&3);
}

/* ----------------- display -------------------------- */


void fen_aide(int mode_printer)
{
  window(1,1,ncols,nlines); /* 25 */
  clrscr();
  window(1,1,ncols,nlines-7); /* 18 */
  gotoxy(1,2);
  textbackground(BLUE);
  textcolor(WHITE);
  /*       1234567890123456789012345678901234567890 */
  gotoxy(1,1);

  cprintf("  Emu71: HP-71B & HP-IL system emulator");
  if (ncols>40)
    cprintf("         J-F GARNIER 1996, 2016         ");
  gotoxy(1,2);
  cprintf(USER);

  /*       1234567890123456789012345678901234567890 */
  window(1,nlines-2,ncols,nlines); /*  window(1,23,80,25); */
  textattr(WHITE);
  clrscr();
  window(1,1,ncols,nlines);  /*   window(1,1,80,25); */
  textattr(LIGHTGRAY<<4);

  gotoxy( 2,nlines-3); cprintf(" f1  ");
  gotoxy( 8,nlines-3); cprintf(" f2  ");
  gotoxy(14,nlines-3); cprintf(" f3  ");
  gotoxy(20,nlines-3); cprintf(" f4  ");

  gotoxy(29,nlines-3); cprintf(" f5  ");
  gotoxy(35,nlines-3); cprintf(" f6  ");
  if (ncols>40) {
    gotoxy(41,nlines-3); cprintf(" f7  ");
    gotoxy(47,nlines-3); cprintf(" f8  ");

    gotoxy(56,nlines-3); cprintf(" f9  ");
    gotoxy(62,nlines-3); cprintf(" f10 ");
    gotoxy(68,nlines-3); cprintf(" f11 ");
    gotoxy(74,nlines-3); cprintf(" f12 ");
  }
  gotoxy( 2,nlines-2); cprintf(" OFF ");
  gotoxy( 8,nlines-2); cprintf("CONT ");
  gotoxy(14,nlines-2); cprintf(" SST ");
  gotoxy(20,nlines-2); cprintf("1USER");

  gotoxy(29,nlines-2); cprintf("FETCH");
  gotoxy(35,nlines-2); cprintf("LIST ");
  if (ncols>40) {
    gotoxy(41,nlines-2); cprintf("DELET");
    gotoxy(47,nlines-2); cprintf("PURGE");

    gotoxy(56,nlines-2); cprintf("INPUT");
    gotoxy(62,nlines-2); cprintf("PRINT");
    gotoxy(68,nlines-2); cprintf("DISP ");
    gotoxy(74,nlines-2); cprintf("AUTO ");
  }
  gotoxy( 2,nlines-1); cprintf("  f  ");
  gotoxy( 8,nlines-1); cprintf("  g  ");
  gotoxy(14,nlines-1); cprintf(" RUN ");
  gotoxy(20,nlines-1); cprintf(" USER");

  gotoxy(29,nlines-1); cprintf("EDIT ");
  gotoxy(35,nlines-1); cprintf(" CAT ");
  if (ncols>40) {
    gotoxy(41,nlines-1); cprintf("NAME ");
    gotoxy(47,nlines-1); cprintf("COPY ");

    gotoxy(56,nlines-1); cprintf("CALL ");
    gotoxy(62,nlines-1); cprintf("GOSUB");
    gotoxy(68,nlines-1); cprintf("RETUR");
    gotoxy(74,nlines-1); cprintf("GOTO ");
  }
  gotoxy(1,nlines);
  textbackground(BLUE);
  textcolor(WHITE);
/*  cprintf("  ESC:ATTN  c-DEL:-LINE  c-ENTER:CMDS  c-E:EDIT  c-F:FETCH  c-T:SST  c-R:RUN "); */
  if (ncols>40)
    cprintf("   Esc:ATTN  ctrl-Del:-LINE  ctrl-Enter:CMDS  Home:|<  End:>|  Pg^:^  PgV:V    ");
  else
    cprintf("Esc:ATTN ctrl-Del:-LINE ctrl-Enter:CMDS");

  textattr(WHITE);

  gotoxy(1,nlines-4);
/*  cprintf("  <-  g  f  AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC  ->        * ");
  cprintf("                     AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC    * ");  */
  gotoxy(1,nlines-5);

  if (mode_printer) cprintf(SIG);

  curx=1;
  cury=2;
}




void prepa_ecran(int mode_printer)
{
  int i;

  gotoxy(39,1);
  cprintf("   ");
  ncols=wherex();  ncols=80;
  gotoxy(1,24);
  cprintf("\n\n\n\n");
  nlines=wherey();
  if (nlines>27) {
    for (i=0;i<8;i++) cprintf("\n\n\n\n");
    nlines=wherey();
  }

  fen_aide(mode_printer);
  fready=1;
  fesc=0;
  adsp=0;
}



void fin_ecran(void)
{
  window(1,1,ncols,nlines);
  textattr(WHITE);
  gotoxy(1,nlines);
  _setcursortype(_NORMALCURSOR);
  cprintf(" \r\n");
  fready=0;
  adsp=0;
}


void LCDwindow(void)
{
  window(1,nlines-5,ncols,nlines-4);
  textattr(WHITE);
}

void SCRwindow(void)
{
  window(1,2,ncols,nlines-7);
  textattr(BLACK*16+LIGHTGREEN);
}




void update_cursor(void)
{
  if ((lcpos!=0)&&(lctyp!=0)) {
    if (adsp!=1) {
      LCDwindow();
      adsp=1;
    }
    gotoxy(lcpos,1);
    if (lctyp==1)
      _setcursortype(_NORMALCURSOR);
    else
      _setcursortype(_SOLIDCURSOR);
  }
  else if (fcur==0) {
    _setcursortype(_NOCURSOR);
  }
  else {
    if (adsp!=2) {
      SCRwindow();
      adsp=2;
    }
    gotoxy(curx, cury);
    if (fctyp==2)
      _setcursortype(_NORMALCURSOR);
    else
    _setcursortype(_SOLIDCURSOR);
  }


}



/* lcd */

void strflags(char *bin, char *ligne)
{
/* "....,....,...x..x.,xx..xxx..xxxx..xxx...x.x.x.x.x..xxx..xxxx..xxxx..xxxx....x. */
/* ".xxxxx.xxxxx.xxxxx.xxxxx....xxxxx.xxxxx.xxxxx.xxxxx....xxxxx.xxxxx.xxxxx.xxxxx */
  strcpy(ligne,"                                                           ");
/*  strcpy(ligne,"              g  f  AC  BAT  USER  RAD  0 1 2 3 4  (o)  PRGM  SUSP  CALC    * "); */

  if (bin[0]&16)  { ligne[ 0]='A'; ligne[ 1]='C'; }                  /* AC */
  if (bin[0]&32)  { ligne[ 9]='U'; ligne[10]='S';
                    ligne[11]='E'; ligne[12]='R'; }                  /* USER */
  if (bin[0]&64)  { ligne[15]='R'; ligne[16]='A'; ligne[17]='D'; }   /* RAD */

/*  if (bin[1]&16)  { ligne[ 4]='B'; ligne[ 5]='A'; ligne[ 6]='T'; }   /* BAT */

  if (bin[2]&8)   { ligne[21]='0'; }                                 /* 0 */
  if (bin[2]&16)  { ligne[23]='1'; }                                 /* 1 */
  if (bin[2]&32)  { ligne[25]='2'; }                                 /* 2 */
  if (bin[2]&64)  { ligne[27]='3'; }                                 /* 3 */
  if (bin[2]&128) { ligne[29]='4'; }                                 /* 4 */

  if (bin[3]&8)   { ligne[30]='('; ligne[31]='o'; ligne[32]=')'; }   /* (o) */
  if (bin[3]&32)  { ligne[37]='P'; ligne[38]='R';
                    ligne[39]='G'; ligne[40]='M'; }                  /* PRGM */
  if (bin[3]&64)  { ligne[43]='S'; ligne[44]='U';
                    ligne[45]='S'; ligne[46]='P'; }                  /* SUSP */
  if (bin[3]&128) { ligne[49]='C'; ligne[50]='A';
                    ligne[51]='L'; ligne[52]='C'; }                  /* CALC */

  if (bin[4])     { ligne[57]='*'; }                                 /* * */
}


void strflags40(char *bin, char *ligne)
{
/*  0123456789012345678901234567890123456789 */
/* ".g.f..USER..RAD.0.1.2.3.4.PRGM..SUSP..*. */
/* ".xxxxx.xxxxx.xxxxx.xxxxx....xxxxx.xxxxx */
  strcpy(ligne, "                                       ");

  if (bin[0]&32)  { ligne[ 6]='U'; ligne[ 7]='S';
		    ligne[ 8]='E'; ligne[ 9]='R'; }                  /* USER */
  if (bin[0]&64)  { ligne[12]='R'; ligne[13]='A'; ligne[14]='D'; }   /* RAD */

  if (bin[2]&8)   { ligne[16]='0'; }                                 /* 0 */
  if (bin[2]&16)  { ligne[18]='1'; }                                 /* 1 */
  if (bin[2]&32)  { ligne[20]='2'; }                                 /* 2 */
  if (bin[2]&64)  { ligne[22]='3'; }                                 /* 3 */
  if (bin[2]&128) { ligne[24]='4'; }                                 /* 4 */

  if (bin[3]&32)  { ligne[26]='P'; ligne[27]='R';
		    ligne[28]='G'; ligne[29]='M'; }                  /* PRGM */
  if (bin[3]&64)  { ligne[32]='S'; ligne[33]='U';
		    ligne[34]='S'; ligne[35]='P'; }                  /* SUSP */

  if (bin[4])     { ligne[38]='*'; }                                 /* * */
}


void update_display(char *buf1, char *buf2, int cpos, int ctype)
{
  char ligne[80];
  char *pt;

  if (adsp!=1) {
    LCDwindow();
    adsp=1;
  }
  gotoxy(1,1);
  if (buf1!=NULL) {
    cputs("                                       ");
    if (ncols>40)
      cputs("                                        ");
    gotoxy(1,1);
    if (ncols<=40) buf1[38]=0;
/* manage the eol in CALC display */
    pt=strchr(buf1,13);
    if (pt) *pt=244;
    cputs(buf1);
  }
  if (buf2!=NULL) {
    if (ncols>40) {
      gotoxy(20,2); /* 22 */
      strflags(buf2,ligne);
      cputs(ligne);
    }
    else {
      gotoxy(7,2);
      strflags40(buf2,ligne);
      cputs(&ligne[6]);
    }
  }

  if (cpos!=0) {
    lcpos=cpos;
    if (ctype&0x40)
      lctyp=0;
    else if (ctype&0x80)
      lctyp=1;
    else
      lctyp=2;
  }
  update_cursor();

}







/* display */



void clrdsp(void)
{

  if (adsp!=2) {
    SCRwindow();
    adsp=2;
  }
  clrscr();
  curx=1;
  cury=1;
  fesc=0;
  fctyp=1;
  fcur=0;

  update_cursor();
}

void disp_string(char *s)
{
  if (!fready)
    printf(s);
  else {
    if (adsp!=2) {
      SCRwindow();
      gotoxy(curx,cury);
      adsp=2;
    }
    cprintf("%s",s);
    curx=wherex();
    cury=wherey();
    update_cursor();
  }
}



void disp_char(int n)
{
  int i;

  if (adsp!=2) {
    SCRwindow();
    gotoxy(curx,cury);
    adsp=2;
  }

  if (fesc) {
    switch (fesc) {
      case 1:
        fesc=0;
        switch (n) {
          case 'A': if (cury>1) cury--; break; /* csrup */
          case 'B': if (cury<(nlines-7)) cury++; break; /* csrdn */
          case 'C':
            if (curx<ncols) curx++;
            else {
              curx=1;
              if (cury<(nlines-7)) cury++;
            }
            break; /* csrr */
          case 'D':
            if (curx>1) curx--;
            else {
              curx=ncols;
              if (cury>1) cury--;
             }
             break; /* csrl */
          case 'E': clrdsp(); break; /* clrdsp */
          case 'H':

            curx=1; cury=1; break; /* home */
          case 'J': /* clrend */
            for (i=curx;i<ncols;i++) cputs(" ");
            gotoxy(curx,cury);
            break;
          case 'Q': fctyp=2; break; /* csrins */
          case 'R': fctyp=1; break; /* csrrep */
          case '<': fcur=0; break; /* csroff */
          case '>': fcur=1; break; /* csron */
          case '%': fesc=2; break; /* csrpos */
        }
        gotoxy(curx,cury);     /* added 1/04/06 for dbg */
        break;
      case 2:
        if (n>=80) n -=64;
        curx=n+1;
        fesc=3;
        break;
      case 3:
        cury=(n&63)+1;
        gotoxy(curx,cury);     /* added 1/04/06 */
        fesc=0;
        break;
    }
  }
  else {
    switch (n) {
      case   0: n='*'; break;
      case  12: n='u'; break;
      case  28: n='s'; break;
      case  27: fesc=1; break;
      case  29: n='#'; break;
/*      case 124: n='a'; break; */
      case 127: n='`'; break;
      case 8:
      case 10:
      case 13: break;
      default:
       if ((n<32)||(n>127))  n='.';   // was removed 08/1/08 for uly2
    }
    if (!fesc) {
      cprintf("%c",n);
      curx=wherex();
      cury=wherey();
    }
  }

  update_cursor();

}


