; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 95
; --------------------------------------------------------------------
; Module sat_arit.asm : mulation des instructions arithmtiques
; --------------------------------------------------------------------
; Version 1.0 16/03/93
;   prototype
; Version 1.1 01/11/95 : modif. traitement dec/hex, macros.
;                        corrections, documentation.
; ********************************************************************


; -------------------------------------------------------------------
; routines arithmtiques :
; x=x+y fs
; x=x-y fs
; x=y-x fs
; x=x+1 fs
; x=x-1 fs
; x=-x  fs
; x=-x-1 fs
;
; addition
; soustraction dans l'un ou l'autre
; incrementation
; decrementation
; negation
; complmentation
;
; mode dcimal ou hexa
; ajuste Carry
; -------------------------------------------------------------------


; ***************************************************************
; 		dfinitions des macros
; ***************************************************************


; -------------------------------------------------------------------
; build_arith2
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _add, _sub, _subr
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------

build_arith2	macro INSTRUCT
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
                local lbl_wp1, lbl_wp2
		local tab_arith2h_wp, tab_arith2d_wp
		local lbl_xs1, lbl_xs2
                local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
                local lbl_m1, lbl_m2
                local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		test_decimal
		js    lbl_p1
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:         call  &INSTRUCT&d_lonib       ; !
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:         call  &INSTRUCT&d_hinib       ; !
lbl_p8:		update_carry
lbl_p9:         pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
                jns   lbl_wp1
                add   bx,32
lbl_wp1:	call  cs:[bx+tab_arith2h_wp]
                pop   di
		maj_a
		pop   si
		NEXTI
tab_arith2h_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
tab_arith2d_wp	dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		test_decimal
		js    lbl_xs1
		call  &INSTRUCT&h_lonib
		pop   si
		NEXTI
lbl_xs1:	call  &INSTRUCT&d_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		call  &INSTRUCT&h_3
		pop   si
		NEXTI
lbl_x1:		call  &INSTRUCT&d_3
		pop   si
		NEXTI

comment !
&INSTRUCT&_s:   add   si,7
		add   di,7
		test_decimal
		js    lbl_s1
		call  &INSTRUCT&h_hinib
		jmp   short lbl_s2
lbl_s1:		call  &INSTRUCT&d_hinib
lbl_s2:		update_carry
		pop   si
		NEXTI
!

&INSTRUCT&_m:
		push  di
		call  maj2_w
		inc   si
		inc   di
		test_decimal
                js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   si
		inc   di
		call  &INSTRUCT&h_11
                jmp   short lbl_m2
lbl_m1:		call  &INSTRUCT&d_hinib
		inc   si
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		pop   si
		NEXTI

		endm







; ===============================================================


; **************************************************************
; 		sub
; dest = dest - src
; **************************************************************

; fonctions de soustraction hexa

subh_byte       macro
		mov  al,[si]
		sbb  [di],al
		inc  si
		inc  di
		endm

_subh_hinib:    mov  al,[si]
		and  al,0f0h
		sub  [di],al
		ret

_subh_16:       subh_byte
_subh_14:       subh_byte
_subh_12:       subh_byte
_subh_10:       subh_byte
_subh_8:        subh_byte
_subh_6:        subh_byte
_subh_4:        subh_byte
_subh_2:        subh_byte
		update_carry
		ret

_subh_15:       subh_byte
_subh_13:       subh_byte
_subh_11:       subh_byte
_subh_9:        subh_byte
_subh_7:        subh_byte
_subh_5:        subh_byte
_subh_3:        subh_byte
_subh_1:
_subh_lonib:    mov  al,[di]
		mov  bh,al
		sbb  al,[si]
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; fonctions de soustraction dcimal

subd_byte       macro
		mov  al,[di]
		sbb  al,[si]
		das
		mov  [di],al
		inc  si
		inc  di
		endm

_subd_hinib:	;
		mov  al,[di]
		mov  ah,[si]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  subd_hi1
		das
subd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_subd_16:       subd_byte
_subd_14:       subd_byte
_subd_12:       subd_byte
_subd_10:       subd_byte
_subd_8:        subd_byte
_subd_6:        subd_byte
_subd_4:        subd_byte
_subd_2:        subd_byte
		update_carry
		ret

_subd_15:       subd_byte
_subd_13:       subd_byte
_subd_11:       subd_byte
_subd_9:        subd_byte
_subd_7:        subd_byte
_subd_5:        subd_byte
_subd_3:        subd_byte
_subd_1:
_subd_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,[si]
		lahf
		and  ah,16
		je   subd_lo1
		sahf
		das
;		update_halfcarry
subd_lo1:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret




; ----------------------------------------


                align 2
_sub_A_B_a:     exec_opcode2s _sub A B a
_sub_B_C_a:     exec_opcode2s _sub B C a
_sub_C_A_a:     exec_opcode2s _sub C A a
_sub_D_C_a:     exec_opcode2s _sub D C a
_sub_B_A_a:     exec_opcode2s _sub B A a
_sub_C_B_a:     exec_opcode2s _sub C B a
_sub_A_C_a:     exec_opcode2s _sub A C a
_sub_C_D_a:     exec_opcode2s _sub C D a

_sub_a:
		test_decimal
		js   _subd_a
_subh_a:        mov  ax,[bx]
		sub  [di],ax
		mov  al,[bx-1]
		sbb  [di-1],al
		update_halfcarry
		NEXT
_subd_a:	mov  ax,[di]
		mov  dx,[bx]
		sub  al,dl
		das
		xchg al,ah
		sbb  al,dh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		sbb  al,[bx-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT



                align 2
_sub_A_B_b:     exec_opcode2s _sub A B b
_sub_B_C_b:     exec_opcode2s _sub B C b
_sub_C_A_b:     exec_opcode2s _sub C A b
_sub_D_C_b:     exec_opcode2s _sub D C b
_sub_B_A_b:     exec_opcode2s _sub B A b
_sub_C_B_b:     exec_opcode2s _sub C B b
_sub_A_C_b:     exec_opcode2s _sub A C b
_sub_C_D_b:     exec_opcode2s _sub C D b

_sub_b:
		test_decimal
		js   _subd_b
_subh_b:        mov  al,[bx]
		sub  [di],al
		update_carry
		NEXTI
_subd_b:	mov  al,[di]
		sub  al,[bx]
		das
		mov  [di],al
		update_carry
		NEXTI



_sub_w:         call maj2_w
		test_decimal
		js   _subd_w
_subh_w:        mov  ax,[si]
		sub  [di],ax
		mov  ax,[si+2]
		sbb  [di+2],ax
		mov  ax,[si+4]
		sbb  [di+4],ax
		mov  ax,[si+6]
		sbb  [di+6],ax
		update_carry
		maj_a
                pop  si
		NEXTI

_subd_w:
                mov  ax,[di]
		mov  bx,[si]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,[di+2]
		mov  bx,[si+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		mov  bx,[si+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		mov  bx,[si+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI


                align 2
_sub_A_B_s:     exec_opcode2s _sub A B s
_sub_B_C_s:     exec_opcode2s _sub B C s
_sub_C_A_s:     exec_opcode2s _sub C A s
_sub_D_C_s:     exec_opcode2s _sub D C s
_sub_B_A_s:     exec_opcode2s _sub B A s
_sub_C_B_s:     exec_opcode2s _sub C B s
_sub_A_C_s:     exec_opcode2s _sub A C s
_sub_C_D_s:     exec_opcode2s _sub C D s

_sub_s:
		test_decimal
		js   _subd_s
_subh_s:    	mov  al,[bx+7]
		and  al,0f0h
		sub  [di+7],al
		update_carry
		NEXTI
_subd_s:	mov  al,[di+7]
		mov  ah,[bx+7]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque   is it needed due to das
		sub  al,ah
		jnc  subd_s1
		das
subd_s1:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


; dfinitions du reste du code
		build_arith2 _sub



; ----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_sub_A_B_&CHAMP&:     exec_opcode2 _sub A B CHAMP
_sub_B_C_&CHAMP&:     exec_opcode2 _sub B C CHAMP
_sub_C_A_&CHAMP&:     exec_opcode2 _sub C A CHAMP
_sub_D_C_&CHAMP&:     exec_opcode2 _sub D C CHAMP
_sub_B_A_&CHAMP&:     exec_opcode2 _sub B A CHAMP
_sub_C_B_&CHAMP&:     exec_opcode2 _sub C B CHAMP
_sub_A_C_&CHAMP&:     exec_opcode2 _sub A C CHAMP
_sub_C_D_&CHAMP&:     exec_opcode2 _sub C D CHAMP

		endm




; **************************************************************
; 		subr
; dest = src - dest
; **************************************************************

; fonctions de soustraction inverse hexa

subrh_byte      macro
		mov  al,[si]
		sbb  al,[di]
		mov  [di],al
		inc  si
		inc  di
		endm

_subrh_hinib:	;
		mov  al,[si]
		mov  ah,[di]
                mov  dh,ah
		and  ax,0f0f0h
                and  dh,00fh
                or   al,dh
		sub  al,ah
		mov  [di],al
		ret

_subrh_16:      subrh_byte
_subrh_14:      subrh_byte
_subrh_12:      subrh_byte
_subrh_10:      subrh_byte
_subrh_8:       subrh_byte
_subrh_6:       subrh_byte
_subrh_4:       subrh_byte
_subrh_2:       subrh_byte
		update_carry
		ret

_subrh_15:      subrh_byte
_subrh_13:      subrh_byte
_subrh_11:      subrh_byte
_subrh_9:       subrh_byte
_subrh_7:       subrh_byte
_subrh_5:       subrh_byte
_subrh_3:       subrh_byte
_subrh_1:
_subrh_lonib:
		mov  al,[si]
		mov  dh,[di]
		sbb  al,dh
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret



; fonctions de soustraction inverse dcimal

subrd_byte      macro
		mov  al,[si]
		sbb  al,[di]
		das
		mov  [di],al
		inc  si
		inc  di
		endm

_subrd_hinib:	;
		mov  al,[si]
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  subrd_hi1
		das
subrd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_subrd_16:      subrd_byte
_subrd_14:      subrd_byte
_subrd_12:      subrd_byte
_subrd_10:      subrd_byte
_subrd_8:       subrd_byte
_subrd_6:       subrd_byte
_subrd_4:       subrd_byte
_subrd_2:       subrd_byte
		update_carry
		ret

_subrd_15:      subrd_byte
_subrd_13:      subrd_byte
_subrd_11:      subrd_byte
_subrd_9:       subrd_byte
_subrd_7:       subrd_byte
_subrd_5:       subrd_byte
_subrd_3:       subrd_byte
_subrd_1:
_subrd_lonib:
		mov  al,[si]
		mov  dh,[di]
		sbb  al,dh
		lahf
		and  ah,16
		je   subrd_lo1
		sahf
		das
;		update_halfcarry
subrd_lo1:	sto_carry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; ----------------------------------------

		align 2

_subr_A_B_a:     exec_opcode2s _subr A B a
_subr_B_C_a:     exec_opcode2s _subr B C a
_subr_C_A_a:     exec_opcode2s _subr C A a
_subr_D_C_a:     exec_opcode2s _subr D C a

_subr_a:
		test_decimal
                js   _subrd_a
_subrh_a:       mov  ax,[bx]
		sub  ax,[di]
		mov  [di],ax
		mov  al,[bx-1]
		sbb  al,[di-1]
		mov  [di-1],al
		update_halfcarry
		NEXT

_subrd_a:       mov  ax,[bx]
		mov  dx,[di]
		sub  al,dl
		das
		xchg al,ah
		sbb  al,dh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[bx-1]
		sbb  al,[di-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT


		align 2

_subr_A_B_b:     exec_opcode2s _subr A B b
_subr_B_C_b:     exec_opcode2s _subr B C b
_subr_C_A_b:     exec_opcode2s _subr C A b
_subr_D_C_b:     exec_opcode2s _subr D C b

_subr_b:
		test_decimal
		js   _subrd_b
_subrh_b:	mov  al,[bx]
		sub  al,[di]
		mov  [di],al
		update_carry
		NEXTI
_subrd_b:	mov  al,[bx]
		sub  al,[di]
		das
		mov  [di],al
		update_carry
		NEXTI

_subr_w:        call maj2_w
		test_decimal
		js   _subrd_w
_subrh_w:       mov  ax,[si]
		sub  ax,[di]
		mov  [di],ax
		mov  ax,[si+2]
		sbb  ax,[di+2]
		mov  [di+2],ax
		mov  ax,[si+4]
		sbb  ax,[di+4]
		mov  [di+4],ax
		mov  ax,[si+6]
		sbb  ax,[di+6]
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

_subrd_w:       mov  ax,[si]
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,[si+2]
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[si+4]
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[si+6]
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

		align 2

_subr_A_B_s:     exec_opcode2s _subr A B s
_subr_B_C_s:     exec_opcode2s _subr B C s
_subr_C_A_s:     exec_opcode2s _subr C A s
_subr_D_C_s:     exec_opcode2s _subr D C s

_subr_s:
		test_decimal
		js   _subrd_s
_subrh_s:
		mov  al,[bx+7]
		mov  ah,[di+7]
		mov  dh,ah
		and  ax,0f0f0h
		and  dh,00fh
		or   al,dh
		sub  al,ah
		mov  [di+7],al
		update_carry
		NEXTI
_subrd_s:	mov  al,[bx+7]
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ax,0f0f0h  ; masque
		sub  al,ah
		jnc  _subrd_s1
		das
_subrd_s1:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


; dfinitions du reste du code
		build_arith2 _subr



; ----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_subr_A_B_&CHAMP&:     exec_opcode2 _subr A B CHAMP
_subr_B_C_&CHAMP&:     exec_opcode2 _subr B C CHAMP
_subr_C_A_&CHAMP&:     exec_opcode2 _subr C A CHAMP
_subr_D_C_&CHAMP&:     exec_opcode2 _subr D C CHAMP

		endm






; **************************************************************
; 		add
; **************************************************************

; fonctions d'addition hexa


addh_byte       macro
		mov  al,[si]
		adc  [di],al
		inc  si
		inc  di
		endm

_addh_hinib:
		mov  al,[si]
		and  al,0f0h
		add  [di],al
		ret



_addh_16:       addh_byte
_addh_14:       addh_byte
_addh_12:       addh_byte
_addh_10:       addh_byte
_addh_8:        addh_byte
_addh_6:        addh_byte
_addh_4:        addh_byte
_addh_2:        addh_byte
		update_carry
		ret



_addh_15:       addh_byte
_addh_13:       addh_byte
_addh_11:       addh_byte
_addh_9:        addh_byte
_addh_7:        addh_byte
_addh_5:        addh_byte
_addh_3:        addh_byte
_addh_1:
_addh_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,[si]
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions d'addition dcimal

addd_byte       macro
		mov  al,[di]
		adc  al,[si]
		daa
		mov  [di],al
		inc  si
		inc  di
		endm

_addd_hinib:	;
		mov  al,[di]
		mov  ah,[si]
                mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		add  al,ah
		daa
                lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_addd_16:       addd_byte
_addd_14:       addd_byte
_addd_12:       addd_byte
_addd_10:       addd_byte
_addd_8:        addd_byte
_addd_6:        addd_byte
_addd_4:        addd_byte
_addd_2:        addd_byte
		update_carry
		ret

_addd_15:       addd_byte
_addd_13:       addd_byte
_addd_11:       addd_byte
_addd_9:        addd_byte
_addd_7:        addd_byte
_addd_5:        addd_byte
_addd_3:        addd_byte
_addd_1:
_addd_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,[si]
		daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; ----------------------------------------

		align 2

_add_A_B_a:     exec_opcode2s _add A B a
_add_B_C_a:     exec_opcode2s _add B C a
_add_C_A_a:     exec_opcode2s _add C A a
_add_D_C_a:     exec_opcode2s _add D C a
_add_A_A_a:     exec_opcode2s _add A A a
_add_B_B_a:     exec_opcode2s _add B B a
_add_C_C_a:     exec_opcode2s _add C C a
_add_D_D_a:     exec_opcode2s _add D D a
_add_B_A_a:     exec_opcode2s _add B A a
_add_C_B_a:     exec_opcode2s _add C B a
_add_A_C_a:     exec_opcode2s _add A C a
_add_C_D_a:     exec_opcode2s _add C D a

_add_a:
		test_decimal
		js   _addd_a
_addh_a:        mov  ax,[bx]
		add  [di],ax
		mov  al,[bx-1]
		adc  [di-1],al
		update_halfcarry
		NEXT

_addd_a:        mov  ax,[di]
		mov  dx,[bx]
		add  al,dl
		daa
		xchg al,ah
		adc  al,dh
		daa
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		adc  al,[bx-1]
		daa
		mov  [di-1],al
		update_halfcarry
		NEXT



                align 2

_add_A_B_b:     exec_opcode2s _add A B b
_add_B_C_b:     exec_opcode2s _add B C b
_add_C_A_b:     exec_opcode2s _add C A b
_add_D_C_b:     exec_opcode2s _add D C b
_add_A_A_b:     exec_opcode2s _add A A b
_add_B_B_b:     exec_opcode2s _add B B b
_add_C_C_b:     exec_opcode2s _add C C b
_add_D_D_b:     exec_opcode2s _add D D b
_add_B_A_b:     exec_opcode2s _add B A b
_add_C_B_b:     exec_opcode2s _add C B b
_add_A_C_b:     exec_opcode2s _add A C b
_add_C_D_b:     exec_opcode2s _add C D b


_add_b:
		test_decimal
		js   _addd_b
_addh_b:	mov  al,[bx]
		add  [di],al
		update_carry
		NEXTI
_addd_b:	mov  al,[di]
		add  al,[bx]
		daa
		mov  [di],al
		update_carry
		NEXTI


_add_w:         call maj2_w
		test_decimal
		js   _addd_w
_addh_w:        mov  ax,[si]
		add  [di],ax
		mov  ax,[si+2]
		adc  [di+2],ax
		mov  ax,[si+4]
		adc  [di+4],ax
		mov  ax,[si+6]
		adc  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

_addd_w:        mov  ax,[di]
		mov  bx,[si]
		add  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di],ax
		mov  ax,[di+2]
		mov  bx,[si+2]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		mov  bx,[si+4]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		mov  bx,[si+6]
		adc  al,bl
		daa
		xchg al,ah
		adc  al,bh
		daa
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		pop  si
		NEXTI

		align 2
_add_A_B_s:     exec_opcode2s _add A B s
_add_B_C_s:     exec_opcode2s _add B C s
_add_C_A_s:     exec_opcode2s _add C A s
_add_D_C_s:     exec_opcode2s _add D C s
_add_A_A_s:     exec_opcode2s _add A A s
_add_B_B_s:     exec_opcode2s _add B B s
_add_C_C_s:     exec_opcode2s _add C C s
_add_D_D_s:     exec_opcode2s _add D D s
_add_B_A_s:     exec_opcode2s _add B A s
_add_C_B_s:     exec_opcode2s _add C B s
_add_A_C_s:     exec_opcode2s _add A C s
_add_C_D_s:     exec_opcode2s _add C D s

_add_s:
		test_decimal
		js   _addd_s
_addh_s:    	mov  al,[bx+7]
		and  al,0f0h
		add  [di+7],al
		update_carry
		NEXTI
_addd_s:	mov  al,[di+7]
		mov  ah,[bx+7]
		mov  dl,al      ; copie original destination
		and  ax,0f0f0h  ; masque
		add  al,ah
		daa
                update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI




; dfinitions du reste du code
		build_arith2 _add

; -----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m, w>

_add_A_B_&CHAMP&:     exec_opcode2 _add A B CHAMP
_add_B_C_&CHAMP&:     exec_opcode2 _add B C CHAMP
_add_C_A_&CHAMP&:     exec_opcode2 _add C A CHAMP
_add_D_C_&CHAMP&:     exec_opcode2 _add D C CHAMP
_add_A_A_&CHAMP&:     exec_opcode2 _add A A CHAMP
_add_B_B_&CHAMP&:     exec_opcode2 _add B B CHAMP
_add_C_C_&CHAMP&:     exec_opcode2 _add C C CHAMP
_add_D_D_&CHAMP&:     exec_opcode2 _add D D CHAMP
_add_B_A_&CHAMP&:     exec_opcode2 _add B A CHAMP
_add_C_B_&CHAMP&:     exec_opcode2 _add C B CHAMP
_add_A_C_&CHAMP&:     exec_opcode2 _add A C CHAMP
_add_C_D_&CHAMP&:     exec_opcode2 _add C D CHAMP

		endm












; ***************************************************************
; 		dfinitions des macros
; ***************************************************************

; -------------------------------------------------------------------
; build_arith
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _inc, _dec, _neg
;          ADJUST_C = stc ou clc selon instruction
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_arith	macro INSTRUCT, ADJUST_C
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
		local lbl_wp1, lbl_wp2
		local tab_arith_wp
		local lbl_xs1, lbl_xs2
		local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
		local lbl_m1, lbl_m2
		local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   di,ax
		test_decimal
		js    lbl_p1
		ADJUST_C
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:		ADJUST_C
		call  &INSTRUCT&d_lonib            ; !
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:		call  &INSTRUCT&d_hinib             ; !
lbl_p8:		update_carry
lbl_p9:         pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
		jns   lbl_wp1
		add   bx,32
lbl_wp1:	ADJUST_C
		call  cs:[bx+tab_arith_wp]
		pop   di
		maj_a
		NEXTI
tab_arith_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
		dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   di
		test_decimal
		js    lbl_xs1
		ADJUST_C
		call  &INSTRUCT&h_lonib
		NEXTI
lbl_xs1:        ADJUST_C
		call  &INSTRUCT&d_lonib
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		ADJUST_C
		call  &INSTRUCT&h_3
		NEXTI
lbl_x1:         ADJUST_C
		call  &INSTRUCT&d_3
		NEXTI

;&INSTRUCT&_s:
;		add   di,7
;		test_decimal
;		js    lbl_s1
;		call  &INSTRUCT&h_hinib
;		update_carry
;		NEXTI
;lbl_s1:         call  &INSTRUCT&d_hinib       ; !
;		update_carry
;		NEXTI

&INSTRUCT&_m:
		push  di
		call  maj_w
		inc   di
		test_decimal
		js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   di
		call  &INSTRUCT&h_11
		jmp   short lbl_m2
lbl_m1:         call  &INSTRUCT&d_hinib
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		NEXTI


		endm





; **************************************************************
; 		inc
; **************************************************************

; fonctions d'incrmentation hexa


inch_byte       macro
		adc  byte ptr [di],0
		inc  di
		endm

_inch_hinib:
		add  byte ptr [di],010h
		ret

_inch_16:       inch_byte
_inch_14:       inch_byte
_inch_12:       inch_byte
_inch_10:       inch_byte
_inch_8:        inch_byte
_inch_6:        inch_byte
_inch_4:        inch_byte
_inch_2:        inch_byte
		update_carry
		ret

_inch_15:       inch_byte
_inch_13:       inch_byte
_inch_11:       inch_byte
_inch_9:        inch_byte
_inch_7:        inch_byte
_inch_5:        inch_byte
_inch_3:        inch_byte
_inch_1:
_inch_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,0
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions d'incrmentation dcimal


incd_byte       macro
		mov  al,[di]
		adc  al,0
		daa
		mov  [di],al
		inc  di
		endm

_incd_hinib:
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_incd_16:       incd_byte
_incd_14:       incd_byte
_incd_12:       incd_byte
_incd_10:       incd_byte
_incd_8:        incd_byte
_incd_6:        incd_byte
_incd_4:        incd_byte
_incd_2:        incd_byte
		update_carry
		ret

_incd_15:       incd_byte
_incd_13:       incd_byte
_incd_11:       incd_byte
_incd_9:        incd_byte
_incd_7:        incd_byte
_incd_5:        incd_byte
_incd_3:        incd_byte
_incd_1:
_incd_lonib:
		mov  al,[di]
		mov  bh,al
		adc  al,0
		daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret

; ----------------------------------


exec_inc_a      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  word ptr _reg&REG&,1
                jc   _inc2
                clear_carry
		NEXT
_inc2:		adc  byte ptr _reg&REG&-1,0
		update_halfcarry
		NEXT
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_a
                endm

_inc_A_a:       exec_inc_a A
_inc_B_a:       exec_inc_a B
_inc_C_a:       exec_inc_a C

_incd_a:        mov  ax,[di]
		add  al,1
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		adc  al,0
		daa
		mov  [di-1],al
		update_halfcarry
		NEXT

_inc_D_a:       exec_inc_a D


exec_inc_b      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  byte ptr _reg&REG&,1
                jc   _inc2
		clear_carry
		NEXTI
_inc2:		set_carry
		NEXTI
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_b
                endm

_inc_A_b:       exec_inc_b A
_inc_B_b:       exec_inc_b B
_inc_C_b:       exec_inc_b C
_inc_D_b:       exec_inc_b D

_incd_b:	mov  al,[di]
		add  al,1
		daa
		mov  [di],al
		update_carry
		NEXTI

                align 2
_inc_A_w:       exec_opcode1s _inc A w
_inc_B_w:       exec_opcode1s _inc B w
_inc_D_w:       exec_opcode1s _inc D w
_inc_C_w:       mov   di, offset _regC

_inc_w:         maj_ws
		test_decimal
		js   _incd_w
_inch_w:        add  word ptr [di],1
                jc   _inch_w1
                clear_carry
		NEXTI
_inch_w1:	adc  word ptr [di+2],0
		adc  word ptr [di+4],0
		adc  word ptr [di+6],0
         	update_carry
		maj_a
		NEXTI
_incd_w:        mov  ax,[di]
		add  al,1
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di],ax
		jnc  _incd_w1            ; oui ?
		mov  ax,[di+2]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		adc  al,0
		daa
		xchg al,ah
		adc  al,0
		daa
		xchg al,ah
		mov  [di+6],ax
_incd_w1:	update_carry
		maj_a
		NEXTI



exec_inc_s      macro REG
                local _incd, _inc2
                align 2
		test_decimal
		js   _incd
                add  byte ptr _reg&REG&+7,010h
                jc  _inc2
		clear_carry
		NEXTI
_inc2:		set_carry
		NEXTI
_incd:          mov  di, offset _reg&REG&
                jmp  short _incd_s
                endm


_inc_A_s:       exec_inc_s A
_inc_B_s:       exec_inc_s B
_inc_C_s:       exec_inc_s C
_inc_D_s:       exec_inc_s D

_incd_s:
		mov  al,[di+7]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI

comment !
                align 2
_inc_A_p:       exec_opcode1s _inc A p
_inc_B_p:       exec_opcode1s _inc B p
_inc_D_p:       exec_opcode1s _inc D p
_inc_C_p:       mov   di, offset _regC

_inc_p: 	test_decimal
		js   _incd_p
                mov  al,_Pointer
                shr  al,1
                jc   _inch_p1
                cmp  al,2
                je   _inch_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		add  al,1
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_inch_p1:       cbw
                add  di,ax
                add  byte ptr [di],010h
		update_carry
                NEXTI
_inch_p4:       add  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_incd_p:        mov  al,_Pointer
                shr  al,1
                jc   _incd_p1
                cmp  al,2
                je   _incd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		add  al,1
                daa
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_incd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		add  al,010h
		daa
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_incd_p4:       mov  al,[di-1]
                add  al,1
                daa
                mov  [di-1],al
                update_halfcarry
                NEXTI
!

; dfinitions du reste du code
		build_arith _inc stc

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_inc_A_&CHAMP&:     exec_opcode _inc A CHAMP
_inc_B_&CHAMP&:     exec_opcode _inc B CHAMP
_inc_C_&CHAMP&:     exec_opcode _inc C CHAMP
_inc_D_&CHAMP&:     exec_opcode _inc D CHAMP

		endm






; -----------------------------------------


; **************************************************************
; 		dec
; **************************************************************

; fonctions de decrmentation hexa


dech_byte       macro
		sbb  byte ptr [di],0
		inc  di
		endm

_dech_hinib:
		sub  byte ptr [di],010h
		ret

_dech_16:       dech_byte
_dech_14:       dech_byte
_dech_12:       dech_byte
_dech_10:       dech_byte
_dech_8:        dech_byte
_dech_6:        dech_byte
_dech_4:        dech_byte
_dech_2:        dech_byte
		update_carry
		ret

_dech_15:       dech_byte
_dech_13:       dech_byte
_dech_11:       dech_byte
_dech_9:        dech_byte
_dech_7:        dech_byte
_dech_5:        dech_byte
_dech_3:        dech_byte
_dech_1:
_dech_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,0
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret



; fonctions de decrmentation dcimal


decd_byte       macro
		mov  al,[di]
		sbb  al,0
		das
		mov  [di],al
		inc  di
		endm

_decd_hinib:
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  decd_hi1     ; pour comportement avec argument > 9 !
		das
decd_hi1:	lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_decd_16:       decd_byte
_decd_14:       decd_byte
_decd_12:       decd_byte
_decd_10:       decd_byte
_decd_8:        decd_byte
_decd_6:        decd_byte
_decd_4:        decd_byte
_decd_2:        decd_byte
		update_carry
		ret

_decd_15:       decd_byte
_decd_13:       decd_byte
_decd_11:       decd_byte
_decd_9:        decd_byte
_decd_7:        decd_byte
_decd_5:        decd_byte
_decd_3:        decd_byte
_decd_1:
_decd_lonib:
		mov  al,[di]
		mov  bh,al
		sbb  al,0
		lahf
		and  ah,16
		je   decd_lo1     ; pour comportement avec argument > 9 !
		sahf
		das
decd_lo1:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


exec_dec_a      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  word ptr _reg&REG&,1
                jc   _dec2
		clear_carry
		NEXT
_dec2:		sbb  byte ptr _reg&REG&-1,0
		update_halfcarry
		NEXT
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_a
                endm


_dec_A_a:       exec_dec_a A
_dec_B_a:       exec_dec_a B
_dec_C_a:       exec_dec_a C

_decd_a:        mov  ax,[di]
		sub  al,1
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di],ax
		mov  al,[di-1]
		sbb  al,0
		das
		mov  [di-1],al
		update_halfcarry
		NEXT

_dec_D_a:       exec_dec_a D



                align 2

exec_dec_b      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  byte ptr _reg&REG&,1
                jc   _dec2
		clear_carry
		NEXTI
_dec2:		set_carry
		NEXTI
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_b
                endm

_dec_A_b:       exec_dec_b A
_dec_B_b:       exec_dec_b B
_dec_C_b:       exec_dec_b C
_dec_D_b:       exec_dec_b D

_decd_b:	mov  al,[di]
                cmp  al,0a0h     ; pour comportement avec digits>9!
                jnc  _dech_b     ; JFG 02/03/03 - added back 18/07/07 !
		sub  al,1
		das
		mov  [di],al
		update_carry
		NEXTI
_dech_b:	sub  byte ptr [di],1
		update_carry
		NEXTI


                align 2
_dec_A_w:       exec_opcode1s _dec A w
_dec_B_w:       exec_opcode1s _dec B w
_dec_D_w:       exec_opcode1s _dec D w
_dec_C_w:       mov   di, offset _regC

_dec_w:         maj_ws
		test_decimal
		js   _decd_w
_dech_w:        sub  word ptr [di],1
                jc   _dech_w1
         	clear_carry
		NEXTI
_dech_w1:       sbb  word ptr [di+2],0
		sbb  word ptr [di+4],0
		sbb  word ptr [di+6],0
         	update_carry
		maj_a
		NEXTI
_decd_w:        mov  ax,[di]
		sub  al,1
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di],ax
		jnc  _decd_w1		; oui ?
		mov  ax,[di+2]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,[di+4]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,[di+6]
		sbb  al,0
		das
		xchg al,ah
		sbb  al,0
		das
		xchg al,ah
		mov  [di+6],ax
_decd_w1:	update_carry
		maj_a
		NEXTI


exec_dec_s      macro REG
                local _decd, _dec2
                align 2
		test_decimal
		js   _decd
                sub  byte ptr _reg&REG&+7,010h
                jc   _dec2
		clear_carry
		NEXTI
_dec2:		set_carry
		NEXTI
_decd:          mov  di, offset _reg&REG&
                jmp  short _decd_s
                endm

_dec_A_s:       exec_dec_s A
_dec_B_s:       exec_dec_s B
_dec_C_s:       exec_dec_s C
_dec_D_s:       exec_dec_s D

_decd_s:	mov  al,[di+7]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _decd_s1     ; pour comportement avec argument > 9 !
		das
_decd_s1:       update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


comment !
                align 2
_dec_A_p:       exec_opcode1s _dec A p
_dec_B_p:       exec_opcode1s _dec B p
_dec_D_p:       exec_opcode1s _dec D p
_dec_C_p:       mov   di, offset _regC

_dec_p: 	test_decimal
		js   _decd_p
                mov  al,_Pointer
                shr  al,1
                jc   _dech_p1
                cmp  al,2
                je   _dech_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		update_halfcarry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_dech_p1:       cbw
                add  di,ax
                sub  byte ptr [di],010h
		update_carry
                NEXTI
_dech_p4:       sub  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_decd_p:        mov  al,_Pointer
                shr  al,1
                jc   _decd_p1
                cmp  al,2
                je   _decd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		lahf
		and  ah,16
		je   _decd_p2     ; pour comportement avec argument > 9
		sahf
		das
_decd_p2:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
        	NEXTI
_decd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _decd_p3     ; pour comportement avec argument > 9
		das
_decd_p3:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_decd_p4:       mov  al,[di-1]
                sub  al,1
		lahf
		and  ah,16
		je   decd_p5     ; pour comportement avec argument > 9
		sahf
                das
_decd_p5:	sto_carry
                mov  [di-1],al
                NEXTI
!




; dfinitions du reste du code
		build_arith _dec stc

; -----------------------------------------








; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_dec_A_&CHAMP&:     exec_opcode _dec A CHAMP
_dec_B_&CHAMP&:     exec_opcode _dec B CHAMP
_dec_C_&CHAMP&:     exec_opcode _dec C CHAMP
_dec_D_&CHAMP&:     exec_opcode _dec D CHAMP

		endm






; **************************************************************
; 		neg
; dest = 0-dest
; **************************************************************

; fonctions de complmentation  2 hexa


negh_byte       macro
		mov  al,0
		sbb  al,[di]
		mov  [di],al
		inc  di
		endm

_negh_hinib:	;
		mov  al,[di]
		mov  ah,al
		and  ax,0f00fh
		sub  al,ah
		mov  [di],al
		ret


_negh_16:       negh_byte
_negh_14:       negh_byte
_negh_12:       negh_byte
_negh_10:       negh_byte
_negh_8:        negh_byte
_negh_6:        negh_byte
_negh_4:        negh_byte
_negh_2:        negh_byte
		update_carry
		ret



_negh_15:       negh_byte
_negh_13:       negh_byte
_negh_11:       negh_byte
_negh_9:        negh_byte
_negh_7:        negh_byte
_negh_5:        negh_byte
_negh_3:        negh_byte
_negh_1:
_negh_lonib:
		mov  al,0
		mov  dh,[di]
		sbb  al,dh
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


; fonctions de complmentation  2 dcimal


negd_byte       macro
		mov  al,0
		sbb  al,[di]
		das
		mov  [di],al
		inc  di
		endm

_negd_hinib:
		mov  al,0
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		lahf
		and  dl,0fh
		or   al,dl
		sahf
		mov  [di],al
		ret

_negd_16:       negd_byte
_negd_14:       negd_byte
_negd_12:       negd_byte
_negd_10:       negd_byte
_negd_8:        negd_byte
_negd_6:        negd_byte
_negd_4:        negd_byte
_negd_2:        negd_byte
		update_carry
		ret

_negd_15:       negd_byte
_negd_13:       negd_byte
_negd_11:       negd_byte
_negd_9:        negd_byte
_negd_7:        negd_byte
_negd_5:        negd_byte
_negd_3:        negd_byte
_negd_1:
_negd_lonib:
		mov  dh,[di]
		mov  al,0
		sbb  al,dh
		das
		update_halfcarry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2

_neg_A_a:     exec_opcode1s _neg A a
_neg_B_a:     exec_opcode1s _neg B a
_neg_D_a:     exec_opcode1s _neg D a
_neg_C_a:     mov   di, offset _regC

_neg_a:
		test_decimal
		js   _negd_a
_negh_a:        xor  ax,ax
		sub  ax,[di]		; neg word ptr [di] ?
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		mov  [di-1],al
		update_halfcarry
		NEXT
_negd_a:        xor  ax,ax
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		das
		mov  [di-1],al
		update_halfcarry
		NEXT

		align 2
_neg_A_b:       exec_opcode1s _neg A b
_neg_B_b:       exec_opcode1s _neg B b
_neg_D_b:       exec_opcode1s _neg D b
_neg_C_b:       mov   di, offset _regC

_neg_b:
		test_decimal
		js   _negd_b
_negh_b:	xor  al,al           ; neg  byte ptr [di] ?
		sub  al,[di]
		mov  [di],al
		update_carry
		NEXTI
_negd_b:	xor  al,al
		sub  al,[di]
		das
		mov  [di],al
		update_carry
		NEXTI


		align 2
_neg_A_w:       exec_opcode1s _neg A w
_neg_B_w:       exec_opcode1s _neg B w
_neg_D_w:       exec_opcode1s _neg D w
_neg_C_w:       mov   di, offset _regC

_neg_w:         maj_ws
		test_decimal
		js   _negd_w
_negh_w:        xor  ax,ax
		sub  ax,[di]
		mov  [di],ax
		mov  ax,0
		sbb  ax,[di+2]
		mov  [di+2],ax
		mov  ax,0
		sbb  ax,[di+4]
		mov  [di+4],ax
		mov  ax,0
		sbb  ax,[di+6]
		mov  [di+6],ax
		update_carry
		maj_a
		NEXTI
_negd_w:        xor  ax,ax
		mov  bx,[di]
		sub  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,0
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,0
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,0
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		update_carry
		maj_a
		NEXTI


		align 2
_neg_A_s:       exec_opcode1s _neg A s
_neg_B_s:       exec_opcode1s _neg B s
_neg_D_s:       exec_opcode1s _neg D s
_neg_C_s:       mov   di, offset _regC

_neg_s:
		test_decimal
		js   _negd_s
_negh_s:	mov  al,[di+7]
		mov  ah,al
		and  ax,0f00fh
		sub  al,ah
		mov  [di+7],al
		update_carry
		NEXTI
_negd_s:
		mov  al,0
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		update_carry
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		NEXTI


comment !
                align 2
_neg_A_p:       exec_opcode1s _neg A p
_neg_B_p:       exec_opcode1s _neg B p
_neg_D_p:       exec_opcode1s _neg D p
_neg_C_p:       mov   di, offset _regC

_neg_p: 	test_decimal
		js   _negd_p
                mov  al,_Pointer
                shr  al,1
                jc   _negh_p1
                cmp  al,2
                je   _negh_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al

		sub  al,1
		update_halfcarry


		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		NEXTI
_negh_p1:       cbw
                add  di,ax
                sub  byte ptr [di],010h
		update_carry
                NEXTI
_negh_p4:       sub  byte ptr [di-1],1
                update_halfcarry
                NEXTI

_negd_p:        mov  al,_Pointer
                shr  al,1
                jc   _negd_p1
                cmp  al,2
                je   _negd_p4
                cbw
                add  di,ax
		mov  al,[di]
		mov  bh,al
		sub  al,1
		lahf
		and  ah,16
		je   _negd_p2     ; pour comportement avec argument > 9
		sahf
		das
_negd_p2:	sto_carry
		mov  ah,bh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
        	NEXTI
_negd_p1:       cbw
                add  di,ax
		mov  al,[di]
		mov  dl,al
		and  al,0f0h
		sub  al,010h
		jnc  _negd_p3     ; pour comportement avec argument > 9
		das
_negd_p3:	update_carry
		and  dl,0fh
		or   al,dl
		mov  [di],al
                NEXTI
_negd_p4:       mov  al,[di-1]
                sub  al,1
		lahf
		and  ah,16
		je   negd_p5     ; pour comportement avec argument > 9
		sahf
                das
_negd_p5:	sto_carry
                mov  [di-1],al
                NEXTI
!




; dfinitions du reste du code
		build_arith _neg clc


; -----------------------------------------




; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_neg_A_&CHAMP&:     exec_opcode _neg A CHAMP
_neg_B_&CHAMP&:     exec_opcode _neg B CHAMP
_neg_C_&CHAMP&:     exec_opcode _neg C CHAMP
_neg_D_&CHAMP&:     exec_opcode _neg D CHAMP

		endm





; **************************************************************
; 		not
; dest = -1-dest
; **************************************************************

; fonctions de complmentation  1



; -------------------------------------------------------------------
; build_not
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _not
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_not	macro INSTRUCT
		local lbl_p1, lbl_p2, lbl_p3, lbl_p8, lbl_p9
		local lbl_wp1, lbl_wp2
		local tab_arith_wp
		local lbl_xs1, lbl_xs2
		local lbl_x1, lbl_x2
		local lbl_s1, lbl_s2
		local lbl_m1, lbl_m2
		local lbl_w1, lbl_w2

&INSTRUCT&_p:
		push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    lbl_p2
		; paire, quartet infrieur
		add   di,ax
		test_decimal
		js    lbl_p1
		call  &INSTRUCT&h_lonib
		jmp   short lbl_p9
lbl_p1:         call  &INSTRUCT&d_lonib
		jmp   short lbl_p9
lbl_p2:         ; impaire, quartet suprieur
		add   di,ax
		test_decimal
		js    lbl_p3
		call  &INSTRUCT&h_hinib
		jmp   short lbl_p8
lbl_p3:         call  &INSTRUCT&d_hinib
lbl_p8:		clear_carry
lbl_p9:         pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:
		push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		test_decimal
		jns   lbl_wp1
		add   bx,32
lbl_wp1:	call  cs:[bx+tab_arith_wp]
		pop   di
		maj_a
		NEXTI
tab_arith_wp	dw    &INSTRUCT&h_1,  &INSTRUCT&h_2,  &INSTRUCT&h_3,  &INSTRUCT&h_4
		dw    &INSTRUCT&h_5,  &INSTRUCT&h_6,  &INSTRUCT&h_7,  &INSTRUCT&h_8
		dw    &INSTRUCT&h_9,  &INSTRUCT&h_10, &INSTRUCT&h_11, &INSTRUCT&h_12
		dw    &INSTRUCT&h_13, &INSTRUCT&h_14, &INSTRUCT&h_15, &INSTRUCT&h_16
		dw    &INSTRUCT&d_1,  &INSTRUCT&d_2,  &INSTRUCT&d_3,  &INSTRUCT&d_4
		dw    &INSTRUCT&d_5,  &INSTRUCT&d_6,  &INSTRUCT&d_7,  &INSTRUCT&d_8
		dw    &INSTRUCT&d_9,  &INSTRUCT&d_10, &INSTRUCT&d_11, &INSTRUCT&d_12
		dw    &INSTRUCT&d_13, &INSTRUCT&d_14, &INSTRUCT&d_15, &INSTRUCT&d_16


&INSTRUCT&_xs:
		inc   di
		test_decimal
		js    lbl_xs1
		call  &INSTRUCT&h_lonib
		NEXTI
lbl_xs1:        call  &INSTRUCT&d_lonib
		NEXTI

&INSTRUCT&_x:
		test_decimal
		js    lbl_x1
		call  &INSTRUCT&h_3
		NEXTI
lbl_x1:		call  &INSTRUCT&d_3
		NEXTI

comment !
&INSTRUCT&_s:
		add   di,7
		test_decimal
		js    lbl_s1
		call  &INSTRUCT&h_hinib
		clear_carry
		NEXTI
lbl_s1:         call  &INSTRUCT&d_hinib
		clear_carry
		NEXTI
!

&INSTRUCT&_m:
		push  di
		call  maj_w
		inc   di
		test_decimal
		js    lbl_m1
		call  &INSTRUCT&h_hinib
		inc   di
		call  &INSTRUCT&h_11
		jmp   short lbl_m2
lbl_m1:         stc
		call  &INSTRUCT&d_hinib
		inc   di
		call  &INSTRUCT&d_11
lbl_m2:         pop   di
		maj_a
		NEXTI

		endm



noth_byte       macro
		not  byte ptr [di]
		inc  di
		endm

_noth_hinib:	xor  byte ptr [di],0f0h
		ret

_noth_16:       noth_byte
_noth_14:       noth_byte
_noth_12:       noth_byte
_noth_10:       noth_byte
_noth_8:        noth_byte
_noth_6:        noth_byte
_noth_4:        noth_byte
_noth_2:        noth_byte
		clear_carry
		ret

_noth_15:       noth_byte
_noth_13:       noth_byte
_noth_11:       noth_byte
_noth_9:        noth_byte
_noth_7:        noth_byte
_noth_5:        noth_byte
_noth_3:        noth_byte
_noth_1:
_noth_lonib:	xor  byte ptr [di],00fh
		clear_carry
		ret


; fonctions de complmentation  1 dcimal


notd_byte       macro
		mov  al,099h
		sub  al,[di]
		das
		mov  [di],al
		inc  di
		endm

_notd_hinib:
		mov  al,090h
		mov  ah,[di]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		and  dl,0fh
		or   al,dl
		mov  [di],al
		ret

_notd_16:       notd_byte
_notd_14:       notd_byte
_notd_12:       notd_byte
_notd_10:       notd_byte
_notd_8:        notd_byte
_notd_6:        notd_byte
_notd_4:        notd_byte
_notd_2:        notd_byte
		clear_carry
		ret

_notd_15:       notd_byte
_notd_13:       notd_byte
_notd_11:       notd_byte
_notd_9:        notd_byte
_notd_7:        notd_byte
_notd_5:        notd_byte
_notd_3:        notd_byte
_notd_1:
_notd_lonib:	mov  dh,[di]
		mov  al,009h
		sub  al,dh
		das
		clear_carry
		mov  ah,dh
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2

_not_A_a:       exec_opcode1s _not A a
_not_B_a:       exec_opcode1s _not B a
_not_D_a:       exec_opcode1s _not D a
_not_C_a:       mov   di, offset _regC

_not_a:
		test_decimal
		js   _notd_a
		not  word ptr [di]
		not  byte ptr [di-1]
		clear_carry
		NEXT
_notd_a:        xor  ax,ax
		stc
		mov  bx,[di]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  al,0
		sbb  al,[di-1]
		das
		mov  [di-1],al
		clear_carry
		NEXT


		align 2
_not_A_b:       exec_opcode1s _not A b
_not_B_b:       exec_opcode1s _not B b
_not_D_b:       exec_opcode1s _not D b
_not_C_b:       mov   di, offset _regC

_not_b:
		test_decimal
		js   _notd_b
		not  byte ptr [di]
		clear_carry
		NEXTI
_notd_b:        xor  ax,ax
		stc
		sbb  al,[di]
		das
		mov  [di],al
		clear_carry
		NEXTI


		align 2
_not_A_w:       exec_opcode1s _not A w
_not_B_w:       exec_opcode1s _not B w
_not_D_w:       exec_opcode1s _not D w
_not_C_w:       mov   di, offset _regC

_not_w:         maj_ws
		test_decimal
		js   _notd_w
		not  word ptr [di]
		not  word ptr [di+2]
		not  word ptr [di+4]
		not  word ptr [di+6]
		clear_carry
		maj_a
		NEXTI
_notd_w:        xor  ax,ax
		stc
		mov  bx,[di]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di],ax
		mov  ax,0
		mov  bx,[di+2]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+2],ax
		mov  ax,0
		mov  bx,[di+4]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+4],ax
		mov  ax,0
		mov  bx,[di+6]
		sbb  al,bl
		das
		xchg al,ah
		sbb  al,bh
		das
		xchg al,ah
		mov  [di+6],ax
		clear_carry
		maj_a
		NEXTI


		align 2
_not_A_s:       exec_opcode1s _not A s
_not_B_s:       exec_opcode1s _not B s
_not_D_s:       exec_opcode1s _not D s
_not_C_s:       mov   di, offset _regC

_not_s:
		test_decimal
		js    _notd_s
            	xor  byte ptr [di+7],0f0h
		clear_carry
		NEXTI
_notd_s:
		mov  al,090h
		mov  ah,[di+7]
		mov  dl,ah      ; copie original destination
		and  ah,0f0h  ; masque
		sub  al,ah
		das
		and  dl,0fh
		or   al,dl
		mov  [di+7],al
		clear_carry
		NEXTI



; dfinitions du reste du code
		build_not _not

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, m>

_not_A_&CHAMP&:     exec_opcode _not A CHAMP
_not_B_&CHAMP&:     exec_opcode _not B CHAMP
_not_C_&CHAMP&:     exec_opcode _not C CHAMP
_not_D_&CHAMP&:     exec_opcode _not D CHAMP

		endm


