; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 2001
; --------------------------------------------------------------------
; Module sat_jmp.inc : mulation des instructions de saut
; --------------------------------------------------------------------
; Version 0.0  04/09/93
;   prototype, faisabilit
; Version 1.0  04/11/95
;   finalisation, mise au point
; Version 2.0  09/09/01
;   add opcodes for HP42S emulation
; ********************************************************************







; goc, gonc:
; de la forme 4aa ou 5aa
; saut en addr+aa, o addr est l'adresse du 2me quartet de l'instruction
; en entre, ah[4-0] contient le 2me quartet
; optimisation:
; goc xy, gonc xy : ici, ax = [ - x  y - ]


; goc xy, gonc xy : ici, ax = [ - y x 5 ]
_goc:           test_carry
                jne  jump
                NEXTI

_gonc:          test_carry
                je   jump
                NEXTI


jump:           shr  ax,4     ; ax = [ - - x y ]
		cbw
		dec  ax    ; dplacement par rapport au 2me quartet

dojump:         ; effectue l'addition entre si,cl et ax
		sar  ax,1
		jnc  dojump1
		xor  cl,4
		jne  dojump1
		inc  si
dojump1:        add  si,ax
		js   fixjump
		NEXT
fixjump:	; changement de segment
		or   ax,ax
		mov  al,_regPC+2
		js   fixminus
		inc  al
		jmp  short fixend
fixminus:	dec  al
fixend:         and  ax,0fh
		mov  _regPC+2,al
		add  ax,ax
		mov  bx,ax
		mov  ax,[_tab_seg+bx]
                or   ax,ax
                je   fixend0
		mov  es,ax
		and  si,07fffh
		NEXT
fixend0:        jmp  resetpc




_goto:
                ; al[7-4]=quartet faible
		mov  bx,es:[si]
		shr  bx,cl         ; bl=quartets moyen et fort
                mov  ah,bl
                sar  ax,4          ; ax: dplacement en quartets
                dec  ax            ; pour pointer sur 2me quartet
                jmp  short dojump


_golong:
		mov  ax,es:[si]
		sar  ax,cl         ; al=octet faible
		mov  bx,es:[si+1]
		sar  bx,cl         ; bl=octet fort
                mov  ah,bl
		jmp  short dojump



_govlng:        mov  ax,es:[si]
                shr  ax,cl         ; al= low nibbles
                inc  si
                mov  dx,es:[si]
                shr  dx,cl         ; dl= medium nibbles, dh=high nibble
		mov  bl,dh
		and  bx,0fh
		mov  _regPC+2,bl
		mov  ah,dl
		xchg ax,si      ; ici bx-si contiennent une adresse 20 bits
_govlng1:
; conversion en pointer es:si, cl
; TEST for jump to address 0 (includes RET with empty stack)
                mov   ax,bx
                or    ax,si
                je    _gobreak
		xor   cl,cl
		shr   si,1
		rcl   cl,3
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                or    bx,bx
                je    _govlng0
		mov   es,bx
		NEXT
_govlng0:	jmp   resetpc
_gobreak:       jmp   invalid


_gosub:
		mov  bl,al         ; bl[7-4]=quartet faible
		fetch_byte         ; al=quartets moyen et fort
                mov  ah,al
                mov  al,bl
		sar  ax,4          ; ax: dplacement en quartets

do_jsr:
		xchg ax,dx         ; save displacement in dx
		; construit bx pointant dans pile retour
                ; build di = stack ptr, decr stack ptr
		mov  al,_retstkptr
		sub  al,4
		mov  _retstkptr,al
		xor  ah,ah
		add  ax,offset _retstk
		xchg ax,di
                ;
		; convert regPC, si, cl in  20 bits addr
		; and save PC
		mov  al,_regPC+2
		mov  [di+2],al
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
		mov  [di],bx
		xchg ax,dx         ; restore displacement
		jmp  dojump


_gosubl:
		fetch_byte
                mov  bl,al         ; octet faible
                fetch_byte
                mov  ah,al         ; octet fort
		mov  al,bl
                jmp  do_jsr


_gosbvl:
		; construit bx pointant dans pile retour
		mov  al,_retstkptr
		sub  al,4
		mov  _retstkptr,al
		xor  ah,ah
		add  ax,offset _retstk
		xchg ax,bx
		; convert regPC, si, cl en adr quartet 20 bits ah-di
		mov  al,cl
		mov  di,si
		shr  al,3   ; carry = bit 0
		rcl  di,1
                add  di,5
		mov  al,_regPC+2
                adc  al,0
		; sauve PC
		mov  [bx],di
		mov  [bx+2],al
                ;
                jmp  _govlng



; rtnc , rtnnc
_rtnc:          test_carry
                jne  _rtn
                NEXTI

_rtnnc:         test_carry
                je   _rtn
                NEXTI


_rtnsc:         set_carry
		jmp  short _rtn

_rtncc:         clear_carry
		jmp  short _rtn

_rti:           ; devrait autoriser interruptions
		jmp  short _rtn

_rtnsxm:
; change JFG 7/01/06: check if call to debugger or breakpoint,
; else just execute normal RTNSXM
		mov  bh,_regPC+2
		mov  bl,cl
		mov  ax,si
                dec  ax      ; pc -= 2
		shr  bl,3   ; carry = bit 0
		rcl  ax,1
                cmp  ax,010h
                jne  _rtnsxm1
                cmp  bh,0fh   ; call to debug?
                je   _rtnsxm8
_rtnsxm1:       cmp  ax,word ptr _lastaddr
                jne  _rtnsxm9
                cmp  bh,byte ptr _lastaddr+2  ; breakpoint?
                jne  _rtnsxm9
_rtnsxm8:       mov  ax,8
		jmp  retour
_rtnsxm9:	or   _hwstatus,1    ; XM
; end of change.
; fall into _rtn

_rtn:           mov  al,_retstkptr
                mov  bl,al
                xor  ah,ah
                add  ax, offset _retstk
		xchg ax,bx
                add  al,4
                mov  _retstkptr,al
                ;
                xor  si,si
		xchg si,[bx]
                xor  ax,ax
		xchg al,[bx+2]
		and  al,0fh
                mov  bx,ax
		mov  _regPC+2,bl
		; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1



_mov_PC_A:      mov  di,offset _regA
                jmp short _mov_PC_x
_mov_PC_C:      mov  di,offset _regC
_mov_PC_x:      mov  si,[di]
                mov  bl,[di-1]
                and  bx,0fh
		mov  _regPC+2,bl
                ; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1


_movi_PC_C:     mov  di,offset _regC
                jmp short _movi_PC_x
                align 2
_movi_PC_A:     mov  di,offset _regA
_movi_PC_x:     mov  si,[di]
                mov  bl,[di-1]
                and  bx,0fh
                ; ici bx-si contiennent une adresse 20 bits
                ; conversion en pointer es:si, cl
		xor   cl,cl
		shr   si,1
		rcl   cl,3
		add   bx,bx
		mov   bx,[_tab_seg+bx]
                or    bx,bx
                je    _movi_PC_0
		mov   es,bx
                jmp  _govlng
_movi_PC_0:     ; cas es=0
                jmp  resetPC


_mov_A_PC:      mov  di,offset _regA
                jmp short _mov_x_PC
_mov_C_PC:      mov  di,offset _regC
_mov_x_PC:
                jmp  invalid
		; convert regPC, si, cl en adr quartet 20 bits ah-bx
		mov  ah,_regPC+2
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
                mov  [di],bx
                mov  [di-1],ah
                NEXT



_xch_A_PC:      mov  di,offset _regA
                jmp short _xch_x_PC
_xch_C_PC:      mov  di,offset _regC
_xch_x_PC:
                jmp  invalid
		; convert regPC, si, cl en adr quartet 20 bits ah-bx
		mov  ah,_regPC+2
		mov  al,cl
		mov  bx,si
		shr  al,3   ; carry = bit 0
		rcl  bx,1
                xchg [di],bx
                xchg [di-1],ah
                mov  si,bx
                mov  bl,ah
                and  bx,0fh
                ; ici bx-si contiennent une adresse 20 bits
                jmp  _govlng1

