; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 95
; --------------------------------------------------------------------
; Module sat_logi.asm : mulation des instructions logiques
; --------------------------------------------------------------------
; Version 1 16/03/93
;   prototype
; version 1.1 04/11/95
;   reste  faire _sr
; ********************************************************************


; routines logiques :
; x=x&y fs
; x=x!y fs

; x=0   fs      *
; x=y   fs      *
; xyex  fs      *

; Rn=x
; x=Rn
; xRnex

; xsl   fs      *
; xsr   fs      *

; xslc
; xsrc
; xsrb

; instructions marked as '*' actually are arith opcodes






; ***************************************************************
; 		dfinitions des macros
; ***************************************************************



; -------------------------------------------------------------------
; build_logic2
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _and, _or
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------


build_logic2	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   si
		NEXT


&INSTRUCT&_wp:  push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
                pop   di
		maj_a
		pop   si
		NEXT
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXT

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXT

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXT

&INSTRUCT&_m:   push  di
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		pop   si
		NEXT


		endm








; **************************************************************
; 		and
; **************************************************************

; fonctions de and binaire


and_byte	macro
		mov  al,[si]
		and  [di],al
		inc  si
		inc  di
		endm

_and_hinib:     mov  al,[si]
		or   al,0fh
		and  [di],al
		ret

_and_16:        and_byte
_and_14:        and_byte
_and_12:        and_byte
_and_10:        and_byte
_and_8:         and_byte
_and_6:         and_byte
_and_4:         and_byte
_and_2:         and_byte
	        ret

_and_15:        and_byte
_and_13:        and_byte
_and_11:        and_byte
_and_9:         and_byte
_and_7:         and_byte
_and_5:         and_byte
_and_3:         and_byte
_and_1:
_and_lonib:     mov  al,[si]
	        or   al,0f0h
	        and  [di],al
	        ret

		align 2
_and_A_B_a:     exec_opcode2s _and A B a
_and_B_C_a:     exec_opcode2s _and B C a
_and_C_A_a:     exec_opcode2s _and C A a
_and_D_C_a:     exec_opcode2s _and D C a
_and_B_A_a:     exec_opcode2s _and B A a
_and_C_B_a:     exec_opcode2s _and C B a
_and_A_C_a:     exec_opcode2s _and A C a
_and_C_D_a:     exec_opcode2s _and C D a

_and_a:         mov  ax,[bx]
		and  [di],ax
		mov  al,[bx-1]
		and  [di-1],al
		NEXT

		align 2
_and_A_B_b:     exec_opcode2s _and A B b
_and_B_C_b:     exec_opcode2s _and B C b
_and_C_A_b:     exec_opcode2s _and C A b
_and_D_C_b:     exec_opcode2s _and D C b
_and_B_A_b:     exec_opcode2s _and B A b
_and_C_B_b:     exec_opcode2s _and C B b
_and_A_C_b:     exec_opcode2s _and A C b
_and_C_D_b:     exec_opcode2s _and C D b

_and_b:         mov  al,[bx]
		and  [di],al
		NEXT


_and_w:         call maj2_w
                mov  ax,[si]
                and  [di],ax
                mov  ax,[si+2]
                and  [di+2],ax
                mov  ax,[si+4]
                and  [di+4],ax
                mov  ax,[si+6]
                and  [di+6],ax
                maj_a
		pop  si
		NEXT

; dfinitions du reste du code
		build_logic2 _and

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_and_A_B_&CHAMP&:     exec_opcode2 _and A B CHAMP
_and_B_C_&CHAMP&:     exec_opcode2 _and B C CHAMP
_and_C_A_&CHAMP&:     exec_opcode2 _and C A CHAMP
_and_D_C_&CHAMP&:     exec_opcode2 _and D C CHAMP
_and_B_A_&CHAMP&:     exec_opcode2 _and B A CHAMP
_and_C_B_&CHAMP&:     exec_opcode2 _and C B CHAMP
_and_A_C_&CHAMP&:     exec_opcode2 _and A C CHAMP
_and_C_D_&CHAMP&:     exec_opcode2 _and C D CHAMP

		endm





; **************************************************************
; 		or
; **************************************************************

; fonctions de or binaire


or_byte		macro
		mov  al,[si]
		or   [di],al
		inc  si
		inc  di
		endm

_or_hinib:      mov  al,[si]
		and  al,0f0h
		or   [di],al
		ret

_or_16:         or_byte
_or_14:         or_byte
_or_12:         or_byte
_or_10:         or_byte
_or_8:          or_byte
_or_6:          or_byte
_or_4:          or_byte
_or_2:          or_byte
	        ret

_or_15:         or_byte
_or_13:         or_byte
_or_11:         or_byte
_or_9:          or_byte
_or_7:          or_byte
_or_5:          or_byte
_or_3:          or_byte
_or_1:
_or_lonib:      mov  al,[si]
		and  al,0fh
		or   [di],al
		ret


		align 2
_or_A_B_a:      exec_opcode2s _or A B a
_or_B_C_a:      exec_opcode2s _or B C a
_or_C_A_a:      exec_opcode2s _or C A a
_or_D_C_a:      exec_opcode2s _or D C a
_or_B_A_a:      exec_opcode2s _or B A a
_or_C_B_a:      exec_opcode2s _or C B a
_or_A_C_a:      exec_opcode2s _or A C a
_or_C_D_a:      exec_opcode2s _or C D a

_or_a:          mov  ax,[bx]
		or  [di],ax
		mov  al,[bx-1]
		or  [di-1],al
		NEXT

		align 2
_or_A_B_b:      exec_opcode2s _or A B b
_or_B_C_b:      exec_opcode2s _or B C b
_or_C_A_b:      exec_opcode2s _or C A b
_or_D_C_b:      exec_opcode2s _or D C b
_or_B_A_b:      exec_opcode2s _or B A b
_or_C_B_b:      exec_opcode2s _or C B b
_or_A_C_b:      exec_opcode2s _or A C b
_or_C_D_b:      exec_opcode2s _or C D b

_or_b:          mov  al,[bx]
		or  [di],al
		NEXT

_or_w:          call maj2_w
                mov  ax,[si]
                or   [di],ax
                mov  ax,[si+2]
                or   [di+2],ax
                mov  ax,[si+4]
                or   [di+4],ax
                mov  ax,[si+6]
                or   [di+6],ax
                maj_a
                pop  si
		NEXT

; dfinitions du reste du code
		build_logic2 _or

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_or_A_B_&CHAMP&:     exec_opcode2 _or A B CHAMP
_or_B_C_&CHAMP&:     exec_opcode2 _or B C CHAMP
_or_C_A_&CHAMP&:     exec_opcode2 _or C A CHAMP
_or_D_C_&CHAMP&:     exec_opcode2 _or D C CHAMP
_or_B_A_&CHAMP&:     exec_opcode2 _or B A CHAMP
_or_C_B_&CHAMP&:     exec_opcode2 _or C B CHAMP
_or_A_C_&CHAMP&:     exec_opcode2 _or A C CHAMP
_or_C_D_&CHAMP&:     exec_opcode2 _or C D CHAMP

		endm









; ========================================================
; copie, change de registres
; ========================================================


; -------------------------------------------------------------------
; build_mov
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  2 arguments
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT =  _mov
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------


build_mov	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
                pop   di
		maj_a
		pop   si
		NEXTI
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXTI

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		pop   si
		NEXTI


		endm





; **************************************************************
; 		mov
; **************************************************************

; fonctions de copie


mov_byte	macro
		mov  al,[si]
		mov  [di],al
		inc  si
		inc  di
		endm

_mov_hinib:     mov  al,[si]
		mov  ah,[di]
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		ret

_mov_16:        mov_byte
_mov_14:        mov_byte
_mov_12:        mov_byte
_mov_10:        mov_byte
_mov_8:         mov_byte
_mov_6:         mov_byte
_mov_4:         mov_byte
_mov_2:         mov_byte
	        ret

_mov_15:        mov_byte
_mov_13:        mov_byte
_mov_11:        mov_byte
_mov_9:         mov_byte
_mov_7:         mov_byte
_mov_5:         mov_byte
_mov_3:         mov_byte
_mov_1:
_mov_lonib:     mov  al,[si]
	        mov  ah,[di]
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		ret


		align 2
_mov_A_B_a:     exec_opcode2s _mov A B a
_mov_B_C_a:     exec_opcode2s _mov B C a
_mov_C_A_a:     exec_opcode2s _mov C A a
_mov_D_C_a:     exec_opcode2s _mov D C a
_mov_B_A_a:     exec_opcode2s _mov B A a
_mov_C_B_a:     exec_opcode2s _mov C B a
_mov_A_C_a:     exec_opcode2s _mov A C a
_mov_C_D_a:     exec_opcode2s _mov C D a

_mov_a:         mov  ax,[bx]
		mov  [di],ax
		mov  al,[bx-1]
		mov  [di-1],al
		NEXT

		align 2
_mov_A_B_b:     exec_opcode2s _mov A B b
_mov_B_C_b:     exec_opcode2s _mov B C b
_mov_C_A_b:     exec_opcode2s _mov C A b
_mov_D_C_b:     exec_opcode2s _mov D C b
_mov_B_A_b:     exec_opcode2s _mov B A b
_mov_C_B_b:     exec_opcode2s _mov C B b
_mov_A_C_b:     exec_opcode2s _mov A C b
_mov_C_D_b:     exec_opcode2s _mov C D b

_mov_b:         mov  al,[bx]
		mov  [di],al
		NEXTI

		align 2
_mov_A_B_w:     exec_opcode2s _mov A B w
_mov_B_C_w:     exec_opcode2s _mov B C w
_mov_C_A_w:     exec_opcode2s _mov C A w
_mov_D_C_w:     exec_opcode2s _mov D C w
_mov_B_A_w:     exec_opcode2s _mov B A w
_mov_C_B_w:     exec_opcode2s _mov C B w
_mov_A_C_w:     exec_opcode2s _mov A C w
_mov_C_D_w:     exec_opcode2s _mov C D w

_mov_w:
; mise  jour source !!
                xchg di,bx
                call maj_w
                xchg di,bx
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
		mov  ax,[bx+6]
                mov  [di+6],ax
                maj_a
                NEXTI

; dfinitions du reste du code
		build_mov _mov


                align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_mov_&REG&_A:	exec_opcode2s _mov1 REG A w
_mov_&REG&_C:	exec_opcode2s _mov1 REG C w
                endm

; copie d'un registre (A,B,C,D) dans Rx
_mov1_w:
; mise  jour source !!
                xchg bx,di
                call maj_w
                xchg bx,di
                ; copie
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
		mov  ax,[bx+6]
		mov  [di+6],ax
                NEXTI

		align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_mov_A_&REG&:	exec_opcode2s _mov2 A REG w
_mov_C_&REG&:	exec_opcode2s _mov2 C REG w
                endm

; copie d'un registre Rx dans (A,B,C,D)
_mov2_w:
                ; copie
                mov  ax,[bx]
                mov  [di],ax
                mov  ax,[bx+2]
                mov  [di+2],ax
                mov  ax,[bx+4]
                mov  [di+4],ax
                mov  ax,[bx+6]
                mov  [di+6],ax
                maj_a
                NEXTI

; new move ?
;               mov  bx,ds
;               xchg es,bx
;               mov  dx,di
;               movsw
;               movsw
;               movsw
;               movsw
;               mov  es,bx
;               mov  di,dx



; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_mov_A_B_&CHAMP&:     exec_opcode2 _mov A B CHAMP
_mov_B_C_&CHAMP&:     exec_opcode2 _mov B C CHAMP
_mov_C_A_&CHAMP&:     exec_opcode2 _mov C A CHAMP
_mov_D_C_&CHAMP&:     exec_opcode2 _mov D C CHAMP
_mov_B_A_&CHAMP&:     exec_opcode2 _mov B A CHAMP
_mov_C_B_&CHAMP&:     exec_opcode2 _mov C B CHAMP
_mov_A_C_&CHAMP&:     exec_opcode2 _mov A C CHAMP
_mov_C_D_&CHAMP&:     exec_opcode2 _mov C D CHAMP

		endm




; -------------------------------------------------------------------
; build_xch
; -------------------------------------------------------------------
; macro de traitement de l'instruction xch
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _xch
; entres: si, di pointent sur les registres oprandes
;          si: source
;          di: destination
;          la sauvegarde de si (pointeur programme) est dans la pile
;          et doit tre restaur avant NEXT
; -------------------------------------------------------------------

build_xch	macro INSTRUCT
		local logic2_p_i, logic2_p_x, tab_logic2_wp

&INSTRUCT&_p:   push  di
		push  si
		call  maj2_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic2_p_i
		; paire, quartet infrieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_lonib
		jmp   short logic2_p_x
logic2_p_i:	; impaire, quartet suprieur
		add   si,ax
		add   di,ax
		call  &INSTRUCT&_hinib
logic2_p_x:     pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI


&INSTRUCT&_wp:  push  di
		push  si
		call  maj2_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		call  cs:[bx+tab_logic2_wp]
		pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI
tab_logic2_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16


&INSTRUCT&_xs:
		inc   si
		inc   di
		call  &INSTRUCT&_lonib
		pop   si
		NEXTI

&INSTRUCT&_x:
		call  &INSTRUCT&_3
		pop   si
		NEXTI

&INSTRUCT&_s:
		add   si,7
		add   di,7
		call  &INSTRUCT&_hinib
		pop   si
		NEXTI

&INSTRUCT&_m:   push  di
		push  si
		call  maj2_w
		inc   si
		inc   di
		call  &INSTRUCT&_hinib
		inc   si
		inc   di
		call  &INSTRUCT&_11
		pop   di
		maj_a
		pop   di
		maj_a
		pop   si
		NEXTI


		endm








; **************************************************************
; 		xch
; **************************************************************

; fonctions de copie


xch_byte	macro
		mov  al,[si]
		xchg al,[di]
		mov  [si],al
		inc  si
		inc  di
		endm

_xch_hinib:	;
		mov  al,[si]
		mov  ah,[di]
		mov  bx,ax
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		mov  ax,bx
		and  ax,0f00fh
		or   al,ah
		mov  [si],al
		ret

_xch_16:        xch_byte
_xch_14:        xch_byte
_xch_12:        xch_byte
_xch_10:        xch_byte
_xch_8:         xch_byte
_xch_6:         xch_byte
_xch_4:         xch_byte
_xch_2:         xch_byte
	        ret

_xch_15:        xch_byte
_xch_13:        xch_byte
_xch_11:        xch_byte
_xch_9:         xch_byte
_xch_7:         xch_byte
_xch_5:         xch_byte
_xch_3:         xch_byte
_xch_1:
_xch_lonib:	mov  al,[si]
		mov  ah,[di]
		mov  bx,ax
		and  ax,0f00fh
		or   al,ah
		mov  [di],al
		mov  ax,bx
		and  ax,0ff0h
		or   al,ah
		mov  [si],al
		ret

		align 2
_xch_A_B_a:     exec_opcode2s _xch A B a
_xch_B_C_a:     exec_opcode2s _xch B C a
_xch_A_C_a:     exec_opcode2s _xch A C a
_xch_C_D_a:     exec_opcode2s _xch C D a

_xch_a:         mov  ax,[bx]
		xchg ax,[di]
		mov  [bx],ax
		mov  al,[bx-1]
		xchg al,[di-1]
		mov  [bx-1],al
		NEXT

		align 2
_xch_A_B_b:     exec_opcode2s _xch A B b
_xch_B_C_b:     exec_opcode2s _xch B C b
_xch_A_C_b:     exec_opcode2s _xch A C b
_xch_C_D_b:     exec_opcode2s _xch C D b

_xch_b:         mov  al,[bx]
		xchg al,[di]
		mov  [bx],al
		NEXTI

_xch_w:         call maj2_w
                mov  ax,[si]
                xchg ax,[di]
                mov  [si],ax
                mov  ax,[si+2]
                xchg ax,[di+2]
                mov  [si+2],ax
                mov  ax,[si+4]
                xchg ax,[di+4]
                mov  [si+4],ax
                mov  ax,[si+6]
                xchg ax,[di+6]
		mov  [si+6],ax
; mise  jour des 2 registres !
		maj_a
		mov  di,si
		maj_a
		pop  si
		NEXTI


; dfinitions du reste du code
		build_xch _xch


                align 2
		irp   REG,<R0, R1, R2, R3, R4, R5, R6, R7>
_xch_A_&REG&:	exec_opcode2s _xch1 A REG w
_xch_C_&REG&:	exec_opcode2s _xch1 C REG w
                endm

; echange entre un registre (A,B,C,D) et Rx
_xch1_w:        call maj_w
		mov  ax,[bx]
		xchg ax,[di]
		mov  [bx],ax
		mov  ax,[bx+2]
		xchg ax,[di+2]
		mov  [bx+2],ax
		mov  ax,[bx+4]
		xchg ax,[di+4]
		mov  [bx+4],ax
		mov  ax,[bx+6]
		xchg ax,[di+6]
		mov  [bx+6],ax
		maj_a
		NEXTI



; -----------------------------------------



; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m, w>

_xch_A_B_&CHAMP&:     exec_opcode2 _xch A B CHAMP
_xch_B_C_&CHAMP&:     exec_opcode2 _xch B C CHAMP
_xch_A_C_&CHAMP&:     exec_opcode2 _xch A C CHAMP
_xch_C_D_&CHAMP&:     exec_opcode2 _xch C D CHAMP

		endm









; -------------------------------------------------------------------
; build_logic
; -------------------------------------------------------------------
; macro de traitement des instructions arithmtiques  1 argument
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m
; argument INSTRUCT = _sl, _clr
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_logic	macro INSTRUCT
		local logic_p_i, logic_p_x, tab_logic_wp

&INSTRUCT&_p:   push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic_p_i
		; paire, quartet infrieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_lonib
		jmp   short logic_p_x
logic_p_i:	; impaire, quartet suprieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_hinib
logic_p_x:      pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj_w
		xor   bx,bx
		mov   bl,_Pointer
		add   bx,bx
		xor   ax,ax
		call  cs:[bx+tab_logic_wp]
		pop   di
		maj_a
		NEXTI
tab_logic_wp	dw    &INSTRUCT&_1,  &INSTRUCT&_2,  &INSTRUCT&_3,  &INSTRUCT&_4
		dw    &INSTRUCT&_5,  &INSTRUCT&_6,  &INSTRUCT&_7,  &INSTRUCT&_8
		dw    &INSTRUCT&_9,  &INSTRUCT&_10, &INSTRUCT&_11, &INSTRUCT&_12
		dw    &INSTRUCT&_13, &INSTRUCT&_14, &INSTRUCT&_15, &INSTRUCT&_16

;&INSTRUCT&_n:   push  di
;		call  maj_w
;		add   bx,bx
;		xor   ax,ax
;		call  cs:[bx+tab_logic_wp]
;		pop   di
;		maj_a
;		NEXT

&INSTRUCT&_xs:  inc   di
		xor   ax,ax
		call  &INSTRUCT&_lonib
		NEXTI

&INSTRUCT&_x:   xor   ax,ax
		call  &INSTRUCT&_3
		NEXTI

&INSTRUCT&_s:   add   di,7
		xor   ax,ax
		call  &INSTRUCT&_hinib
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj_w
		inc   di
                xor   ax,ax
		call  &INSTRUCT&_hinib
		inc   di
		call  &INSTRUCT&_11
                pop   di
		maj_a
		NEXTI

		endm





; **************************************************************
; 		clr
; **************************************************************

; fonctions de mise  0


clr_byte        macro
		mov  [di],al
		inc  di
		endm

_clr_hinib:     and  byte ptr [di],0fh
		ret

_clr_16:        clr_byte
_clr_14:        clr_byte
_clr_12:        clr_byte
_clr_10:        clr_byte
_clr_8:         clr_byte
_clr_6:         clr_byte
_clr_4:         clr_byte
_clr_2:         clr_byte
		ret

_clr_15:        clr_byte
_clr_13:        clr_byte
_clr_11:        clr_byte
_clr_9:         clr_byte
_clr_7:         clr_byte
_clr_5:         clr_byte
_clr_3:         clr_byte
_clr_1:
_clr_lonib:     and  byte ptr [di],0f0h
		ret


		align 2
_clr_A_a:       exec_opcode1s _clr A a
_clr_B_a:       exec_opcode1s _clr B a
_clr_D_a:       exec_opcode1s _clr D a
;_clr_C_a:       exec_opcode1s _clr C a
_clr_C_a:       mov   di, offset _regC

_clr_a:         xor  ax,ax
		mov  [di],ax
		mov  [di-1],al
		NEXT

		align 2
_clr_A_b:       exec_opcode1s _clr A b
_clr_B_b:       exec_opcode1s _clr B b
_clr_D_b:       exec_opcode1s _clr D b
_clr_C_b:       mov   di, offset _regC

_clr_b:		mov  byte ptr [di],0
		NEXTI

		align 2
_clr_A_w:       exec_opcode1s _clr A w
_clr_B_w:       exec_opcode1s _clr B w
_clr_D_w:       exec_opcode1s _clr D w
_clr_C_w:       mov   di, offset _regC

_clr_w:         xor  ax,ax
		mov  [di],ax
		mov  [di+2],ax
		mov  [di+4],ax
		mov  [di+6],ax
		maj_a
		NEXTI

; dfinitions du reste du code
		build_logic _clr

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_clr_A_&CHAMP&:     exec_opcode _clr A CHAMP
_clr_B_&CHAMP&:     exec_opcode _clr B CHAMP
_clr_C_&CHAMP&:     exec_opcode _clr C CHAMP
_clr_D_&CHAMP&:     exec_opcode _clr D CHAMP

		endm
















; **************************************************************
; 		sl
; **************************************************************

; fonctions de dcallage  gauche de 1 quartet
; SB inchang


; ah[7-4] doit contenir le quartet issu du decallage precedent
sl_byte         macro
		mov  al,[di]
		rol  ax,4
		mov  [di],al
		shl  ah,4   ; quartet decalle en ah[7-0]
		inc  di
		endm

_sl_hinib:      mov  al,[di]
		mov  ah,al
		and  al,0fh
		mov  [di],al
		ret

_sl_16:         sl_byte
_sl_14:         sl_byte
_sl_12:         sl_byte
_sl_10:         sl_byte
_sl_8:          sl_byte
_sl_6:          sl_byte
_sl_4:          sl_byte
_sl_2:          sl_byte
		ret

_sl_15:         sl_byte
_sl_13:         sl_byte
_sl_11:         sl_byte
_sl_9:          sl_byte
_sl_7:          sl_byte
_sl_5:          sl_byte
_sl_3:          sl_byte
_sl_1:
_sl_lonib:      mov  al,[di]
		shr  ah,4
		and  ax,0ff0h
		or   al,ah
		mov  [di],al
		ret


		align 2
_sl_A_a:        exec_opcode1s _sl A a
_sl_B_a:        exec_opcode1s _sl B a
_sl_D_a:        exec_opcode1s _sl D a
_sl_C_a:        mov   di, offset _regC

_sl_a:		mov  ax,[di]
		rol  ax,4
		mov  [di-1],al
		and  al,0f0h
		mov  [di],ax
		NEXT

		align 2
_sl_A_b:        exec_opcode1s _sl A b
_sl_B_b:        exec_opcode1s _sl B b
_sl_D_b:        exec_opcode1s _sl D b
_sl_C_b:        mov   di, offset _regC

_sl_b:          mov  al,[di]
		shl  al,4
		mov  [di],al
		NEXTI

		align 2
_sl_A_w:        exec_opcode1s _sl A w
_sl_B_w:        exec_opcode1s _sl B w
_sl_D_w:        exec_opcode1s _sl D w
_sl_C_w:        mov   di, offset _regC

_sl_w:          call maj_w
		push cx
		mov  ax,[di]
		mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shl  ax,1
		rcl  bx,1
		rcl  cx,1
		rcl  dx,1
                endm
                mov  [di],ax
                mov  [di+2],bx
                mov  [di+4],cx
                mov  [di+6],dx
                pop  cx
		maj_a
		NEXTI



; dfinitions du reste du code
		build_logic _sl

; -----------------------------------------






; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_sl_A_&CHAMP&:     exec_opcode _sl A CHAMP
_sl_B_&CHAMP&:     exec_opcode _sl B CHAMP
_sl_C_&CHAMP&:     exec_opcode _sl C CHAMP
_sl_D_&CHAMP&:     exec_opcode _sl D CHAMP

		endm








; -----------------------------------------

; -------------------------------------------------------------------
; build_logic0
; -------------------------------------------------------------------
; macro de traitement des instructions sr
; -------------------------------------------------------------------
; gre les oprations sur les champs p, wp, xs, s, m, w
; argument INSTRUCT = _sr
; entres: di pointe sur le registre oprande destination
; -------------------------------------------------------------------

build_sr	macro INSTRUCT
		local logic_p_i, logic_p_x, tab_logic_wp

&INSTRUCT&_p:   push  di
		call  maj_w
		xor   ah,ah
		mov   al,_Pointer
		shr   al,1
		jc    logic_p_i
		; paire, quartet infrieur
		add   di,ax
		call  &INSTRUCT&_lonib
		call  &INSTRUCT&_0
		jmp   short logic_p_x
logic_p_i:	; impaire, quartet suprieur
		add   di,ax
		xor   ax,ax
		call  &INSTRUCT&_hinib
logic_p_x:      pop   di
		maj_a
		NEXTI


&INSTRUCT&_wp:  push  di
		call  maj_w
		xor   ax,ax
		mov   al,_Pointer
		mov   bx,ax
		shr   al,1
		jc    logic_wp1
		add   di,ax
		call  &INSTRUCT&_lonib
		dec   di
		; ici, bx paire, ok pour jmp
		jmp   short logic_wp2
logic_wp1:      add   di,ax
		xor   ax,ax
		; ajuste bx pour etre paire, et ok pour jmp
		inc   bx
logic_wp2:      call  cs:[bx+tab_logic_wp]
		pop   di
		maj_a
		NEXTI
tab_logic_wp	dw    &INSTRUCT&_0,  &INSTRUCT&_2,  &INSTRUCT&_4,  &INSTRUCT&_6
		dw    &INSTRUCT&_8,  &INSTRUCT&_10, &INSTRUCT&_12, &INSTRUCT&_14
		dw    &INSTRUCT&_16

&INSTRUCT&_xs:  inc   di
		call  &INSTRUCT&_lonib
		call  &INSTRUCT&_0
		NEXTI

&INSTRUCT&_x:   inc   di
		call  &INSTRUCT&_lonib
		dec   di
		call  &INSTRUCT&_2
		NEXTI

&INSTRUCT&_s:   add   di,7
		xor   ax,ax
		call  &INSTRUCT&_hinib
		NEXTI

&INSTRUCT&_m:   push  di
		call  maj_w
		add   di,7
		xor   ax,ax
		call  &INSTRUCT&_lonib
		dec   di
		call  &INSTRUCT&_11
		pop   di
		maj_a
		NEXTI

;&INSTRUCT&_w:   push  di
;		call  maj_w
;                xor   ax,ax
;		call  &INSTRUCT&_16
;                pop   di
;		maj_a
;		NEXTI

		endm





; **************************************************************
; 		sr
; **************************************************************

; fonctions de dcallage  droite de 1 quartet
; SB=1 si quartet perdu<>0

; ah[3-0] doit contenir le quartet issu du decallage precedent
sr_byte         macro
		mov  al,[di]
		ror  ax,4
		mov  [di],al
		dec  di
		shr  ah,4
		endm

_sr_lonib:      mov  al,[di]
		mov  ah,al
		and  al,0f0h
		mov  [di],al
		ret

_sr_16:         sr_byte
_sr_14:         sr_byte
_sr_12:         sr_byte
_sr_10:         sr_byte
_sr_8:          sr_byte
_sr_6:          sr_byte
_sr_4:          sr_byte
_sr_2:          sr_byte
_sr_0:          and  ah,0fh
		je   _sr_00
		or   _hwstatus,2       ; SB
_sr_00:         ret

_sr_15:         sr_byte
_sr_13:         sr_byte
_sr_11:         sr_byte
_sr_9:          sr_byte
_sr_7:          sr_byte
_sr_5:          sr_byte
_sr_3:          sr_byte
_sr_1:
_sr_hinib:      mov  al,[di]
		mov  bh,al
		ror  ax,4
		mov  bl,al
		and  bx,0ff0h
		or   bl,bh
		mov  [di],bl
		and  al,0fh
		je   _sr_h1
		or   _hwstatus,2       ; SB
_sr_h1:	        ret



		align 2
_sr_A_a:        exec_opcode1s _sr A a
_sr_B_a:        exec_opcode1s _sr B a
_sr_D_a:        exec_opcode1s _sr D a
_sr_C_a:        mov   di, offset _regC

_sr_a:          xor  bx,bx
		mov  ax,[di]         ; ax : [ 3 2 1 0 ]
		xchg bl,[di-1]       ; bl : [ x 4 ]
		mov  bh,al           ; bx : [ 1 0 x 4 ]
		and  bx,0f0fh        ; bx = [ - 0 - 4 ]
		or   bh,bh
		je   _sr_a1
		or   _hwstatus,2       ; SB
_sr_a1:         and  al,0f0h         ; ax : [ 3 2 1 - ]
		or   al,bl           ; ax : [ 3 2 1 4 ]
		ror  ax,4            ; ax : [ 4 3 2 1 ]
		mov  [di],ax
		NEXT

		align 2
_sr_A_b:        exec_opcode1s _sr A b
_sr_B_b:        exec_opcode1s _sr B b
_sr_D_b:        exec_opcode1s _sr D b
_sr_C_b:        mov   di, offset _regC

_sr_b:          xor  ax,ax
		mov  al,[di]
		ror  ax,4
		mov  [di],al
		or   ah,ah
		je   _sr_b1
		or   _hwstatus,2       ; SB
_sr_b1:		NEXTI

		align 2
_sr_A_w:        exec_opcode1s _sr A w
_sr_B_w:        exec_opcode1s _sr B w
_sr_D_w:        exec_opcode1s _sr D w
_sr_C_w:        mov   di, offset _regC

_sr_w:          call maj_w
		push cx
		mov  ax,[di]
		mov  bx,ax
		and  bl,0fh
		je   _sr_w1
		or   _hwstatus,2       ; SB
_sr_w1:         mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  cx
		maj_a
		NEXTI


; dfinitions du reste du code
		build_sr _sr

; -----------------------------------------







; dfinitions des points d'entres
		align 2
		irp   CHAMP,<p, wp, xs, x, s, m>

_sr_A_&CHAMP&:     exec_opcode _sr A CHAMP
_sr_B_&CHAMP&:     exec_opcode _sr B CHAMP
_sr_C_&CHAMP&:     exec_opcode _sr C CHAMP
_sr_D_&CHAMP&:     exec_opcode _sr D CHAMP

		endm






; **************************************************************
; rotations
; _slc, _src, _srb
; **************************************************************


		align 2
_slc_A:         exec_opcode1s _slc A w
_slc_B:         exec_opcode1s _slc B w
_slc_D:         exec_opcode1s _slc D w

; rotation d'un quartet vers la gauche du champ w
; SB inchang
_slc_w:
		call maj_w
		push cx
		push si
		mov  dx,[di+6]
		mov  si,dx
		mov  cx,[di+4]
		mov  bx,[di+2]
		mov  ax,[di]
		rept 4
		shl  si,1
		rcl  ax,1
		rcl  bx,1
		rcl  cx,1
		rcl  dx,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  si
		pop  cx
		maj_a
		NEXTI


_slc_C:
		mov  di, offset _regC
                jmp  short _slc_w   ; to be removed
;  this is a code to execute 2 CSLC W consecutive opcodes in one go:
                incr_pc
                mov  ax,es:[si]
                sar  ax,cl
                and  ah,0fh
                cmp  ax,0218h   ; CSRC is 812
                jne  short _slc_w
; ok next opcode is a 2nd CSLC
                inc  si
                maj_ws
		mov  bl,[di+7]
                mov  ax,[di+5]
                mov  [di+6],ax
                mov  ax,[di+3]
                mov  [di+4],ax
                mov  ax,[di+1]
                mov  [di+2],ax
                mov  ah,[di]
                mov  al,bl
                mov  [di],ax
                maj_a
                NEXTI



		align 2
_src_A:         exec_opcode1s _src A w
_src_B:         exec_opcode1s _src B w
_src_D:         exec_opcode1s _src D w

; rotation d'un quartet vers la droite du champ w
; SB=1 si quartet report en poids fort <>0
_src_w:
                call maj_w
                push cx
		push si
		mov  ax,[di]
		mov  si,ax
		mov  bx,[di+2]
		mov  cx,[di+4]
		mov  dx,[di+6]
		rept 4
		shr  si,1
		rcr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		endm
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  si
		pop  cx
		and  dh,0f0h
		je   _src_w1
		or   _hwstatus,2       ; SB
_src_w1:        maj_a
                NEXTI


_src_C:
		mov   di, offset _regC
                jmp   short _src_w   ; to be removed
;  this is a code to execute 2 CSRC W consecutive opcodes in one go:
                incr_pc
                mov ax,es:[si]
                sar ax,cl
                and ah,0fh
                cmp ax,0618h   ; CSRC is 816
                jne short _src_w
; ok next opcode is a 2nd CSRC
                inc si
                maj_ws
		mov  bl,[di]
		or   bl,bl
		je   _sr_w3
		or   _hwstatus,2       ; SB
_sr_w3:
                mov ax,[di+1]
                mov [di],ax
                mov ax,[di+3]
                mov [di+2],ax
                mov ax,[di+5]
                mov [di+4],ax
                mov al,[di+7]
                mov ah,bl
                mov [di+6],ax
                maj_a
                NEXTI





		align 2
_srb_A:         exec_opcode1s _srb A w
_srb_B:         exec_opcode1s _srb B w
_srb_D:         exec_opcode1s _srb D w

_srb_C:		mov   di, offset _regC

; dcallage d'un bit vers la droite du champ w
; SB=1 si carry
_srb_w:
		call maj_w
                push cx
		mov  ax,[di]
                mov  bx,[di+2]
                mov  cx,[di+4]
		mov  dx,[di+6]
		shr  dx,1
		rcr  cx,1
		rcr  bx,1
		rcr  ax,1
		mov  [di],ax
		mov  [di+2],bx
		mov  [di+4],cx
		mov  [di+6],dx
		pop  cx
		jnc  _srb_w1
		or   _hwstatus,2       ; SB
_srb_w1:        maj_a
		NEXTI








