; ********************************************************************
;               mulation du processeur Saturn par le 8086
;                       J-F Garnier 1993, 2004
; --------------------------------------------------------------------
; Module saturn.asm : dfinitions des registres, moteur d'excution
; --------------------------------------------------------------------
; Version 0.0 16/03/93
;   prototype, faisabilit
; Version 1.0 04/11/95
; Version 2.0 07/09/01 extensions pour HP42
; Version 2.1 08/09/02 maj
; Version 3.0 03/10/02 12 bits jump table
; Version 3.1 10/07/07 integration in Emu71
;
; ********************************************************************




		.286


		.model large


		; mulation du processeur Saturn par le 8086
		; J-F Garnier
		; conception 16/03/93
		; Version 1 : support du 1LF2 (hp71)
		;         volutions prvues: 1LK7 (hp28), 1LR2 (hp48)
		;
		; Modif:
		; 04/09/93 : commentaires
		; Reprise 29/10/95: optimisations


		; usage des registres 8086
		; es:si, cl   : simulation du PC.
		; ax, bx, dx  : temporaires.
		; ch          : flags
		; di          : gnralement pointeur registre destination.
		; bp          : compteur d'instructions (_next)


		.data

		; ---------------------------------------
		; REGISTRES DU SATURN



		public _regA, _regB, _regC, _regD
		public _regR0,_regR1,_regR2,_regR3
		public _regR4,_regR5,_regR6,_regR7
		public _regD0,_regD1,_regPC
		public _Carry,_Pointer
		public _retstk,_lastaddr
                public _status,_hwstatus
		public _input,_output
		public _flagdec,_retstkptr
		public _tab_seg
		public _timer_to,_last_ticks
;		public _flagbrk    ; emu42

;                extrn _dsp_offset: byte

; structure des registres d'mulation:
; octet    nibbles
;  7         15,14
;  6         13,12
;  5         11,10
;  4         9, 8
;  3         7, 6
;  2         5, 4
;  1         3, 2
;  0         1, 0
; -1         x, 4      duplication du nibble 4 pour optimiser la vitesse :
; -2         x, x      les instructions sur le champ A utilisent
;                      les octets -1, 0 et 1
;                      (l'octet -2 est rserv pour la version 32 bits)


; ordre des quartets dans les registres
;  -2 -1  0  1  2 ...  7
;  xx x4 10 32 54 ... fe


; organisation mmoire:
;   ordre des quartets en mmoire:
;  octets     0  1  2  3  4 ...
;  quartets  10 32 54 76 98 ...
; pour lire un quartet  l'adresse n:
;   lire octet  adresse n/2                octet = 10
;   si n paire, quartet = octet & 15        quartet = 0
;   si n impaire, quartet = octet >> 4      quartet = 1
; pour lire un octet  l'adresse n:
;   lire mot  l'adresse n/2                mot = 3210
;   si n paire, octet = mot & 255           octet = 10
;   si n impaire, octet = (mot >> 4)&255    octet = 21


                align 2

tabopcode       dw 4096 dup(?)

_retstk         db 256 dup(?)         ; 64 level stack


reg             db 0, 0
_regA           db 10 dup (0)
_regB           db 10 dup (0)
_regC           db 10 dup (0)
_regD           db 10 dup (0)
_regR0          db 10 dup (0)
_regR1          db 10 dup (0)
_regR2          db 10 dup (0)
_regR3          db 10 dup (0)
_regR4          db 10 dup (0)
_regR5          db 10 dup (0)
_regR6          db 10 dup (0)
_regR7          db 10 dup (0)

		; control registers
_Carry          db 0
_Pointer        db 0
_regD0          db 0, 0, 0, 0
_regD1          db 0, 0, 0, 0
_regPC          db 0, 0, 0, 0
_oldretstk         db 16*4 dup(0)         ; pile  8 (16) niveaux
_lastaddr       db 0ffh, 0ffh, 0ffh, 0ffh
_status         dw 0    ; register ST 16 flags
_output         dw 0
_input          dw 0
_hwstatus       db 0    ; bits 3-0 : MP, SR, SB, XM
		db 0

		; internal registers
_flagdec        db 0       ; decimal mode flag
_retstkptr      db 0       ; retstk  pointer
_stackptr       dw 0
cpt_instr	dw 0
cpt_ticks       dw 0       ; compteur de ticks timer 71
_last_ticks     dw 0
_timer_to       db 0
;cpt_h		db 0
;_flagbrk        db 0
tabloaded       db 0

		; ---------------------------------------

_tab_seg        dw 16 dup(0)            ; table of memory segments (32k each)

;cpttimer        dw 0
;lasttic         dw 0
;cpttic          dw 0

; tab semi graphic character for text mode display
;tabsg           db 32, 223, 220, 219
;dspseg          dw 0

;lastbyte        db 0, 0




		.code


		public _executeSaturn

;                public _init_tab

;                public _updatedsp


FASTFLAGS       equ 1



; ==============================================================
; macros de lecture d'opcode
; ==============================================================


; macro de lecture d'un nibble de code
; al[3-0] = (PC); PC += 1;
; Note : al[7-4]=0
; PC est constitu de es:si et cl.
fetch_nib       macro
		local fetch_nib1
		mov  al,es:[si]
		xor  cl,4
		jne  fetch_nib1
		sar  al,4
		inc  si
fetch_nib1:     and  al,15
		endm

incr_pc         macro
		local incr_pc1
		xor  cl,4
		jne  incr_pc1
		inc  si
incr_pc1:
		endm


; macro de lecture d'un octet de code
; Note : ah dtruit
fetch_byte      macro
		mov  ax,es:[si]
		sar  ax,cl      ; al[3-0] = (PC); al[7-4]= (PC+1);
		inc  si         ; PC += 2;
		endm

; macro de lecture de 3 nibbles de code
; en ah[3-0], al
fetch_3nib      macro
		local fetch_3nib1
		mov  ax,es:[si]
		xor  cl,4
		jne  fetch_3nib1
		sar  ax,4
		inc  si
fetch_3nib1:    inc  si
		and  ah,15
		endm



; ***********************************************************
;               s/p & macros gnrales
; ***********************************************************



                IF FASTFLAGS

; flags:
; 7:decimal, 4:half-carry, 2:A-field, 1:pointer, 0:carry

		; met tat carry  0
clear_carry     macro
                and  ch,0eeh
		endm

		; met tat carry  1
set_carry       macro
                or  ch,011h
		endm

sto_carry       macro
                and  ch,0eeh
		or   ch,ah
		endm

test_carry      macro
                test ch,011h
                endm

test_decimal    macro
                or   ch,ch
                endm

set_decimal     macro
                or   ch,080h
                endm

clr_decimal     macro
                and  ch,07fh
                endm

test_pointer    macro
                test ch,2
                endm

set_pointer     macro
                or   ch,2
                endm

clr_pointer     macro
                and  ch,0fdh
                endm

set_majw        macro
;                or   ch,4
                endm

clr_majw        macro
;                and  ch,0fbh
                endm

;test_majw       macro
;                test ch,4
;                endm

                ELSE

		; met tat carry  0
clear_carry     macro
		mov  _Carry,0
		endm

		; met tat carry  0
set_carry       macro
		mov  _Carry,0ffh
		endm

sto_carry       macro
		mov  _Carry,ah
		endm

test_carry      macro
                test _Carry,0ffh
                endm

test_decimal    macro
                test  _flagdec,0ffh
                endm

set_decimal     macro
                mov   _flagdec,0ffh
                endm

clr_decimal     macro
                mov   _flagdec,0
                endm

test_pointer    macro
                test _Pointer,0fh
                endm

set_pointer0    macro
                endm

clr_pointer0    macro
                endm


                ENDIF




		; met  jour tat carry
update_carry    macro
		lahf
		and  ah,1
                sto_carry
		endm

		; met  jour tat half carry
update_halfcarry    macro
		lahf
		and  ah,010h
                sto_carry
		endm


		align 2

		; mettre a jour registre entier (champ W) src et dest
		; src et dest sont points par si et di
maj2_w:		mov   al,[si+2]
		mov   ah,[si-1]
		and   ax,0ff0h
		or    al,ah
		mov   [si+2],al
		; mettre a jour registre entier (champ W) dest
		; dest est point par di
maj_w:          mov   al,[di+2]
		mov   ah,[di-1]
		and   ax,0ff0h
		or    al,ah
		mov   [di+2],al
		ret

maj_ws          macro
                mov   al,[di+2]
		mov   ah,[di-1]
		and   ax,0ff0h
		or    al,ah
		mov   [di+2],al
                endm

maj2_ws         macro
                mov   al,[bx+2]
		mov   ah,[bx-1]
		and   ax,0ff0h
		or    al,ah
		mov   [bx+2],al
                maj_ws
                endm

		; mettre  jour pseudo champ A de dest point par di
maj_a           macro
		mov   al,[di+2]
		mov   [di-1],al
		endm

; macro d'execution d'un opcode

exec_opcode	macro INSTRUCT, DEST, CHAMP
		mov   di, offset _reg&DEST&
		jmp   &INSTRUCT&_&CHAMP&
		endm

exec_opcode2	macro INSTRUCT, DEST, SRC, CHAMP
		push  si
		mov   si, offset _reg&SRC&
		mov   di, offset _reg&DEST&
		jmp   &INSTRUCT&_&CHAMP&
		endm

exec_opcode1s	macro INSTRUCT, DEST, CHAMP
		mov   di, offset _reg&DEST&
		jmp   short &INSTRUCT&_&CHAMP&
		endm

exec_opcode2s	macro INSTRUCT, DEST, SRC, CHAMP
		mov   bx, offset _reg&SRC&
		mov   di, offset _reg&DEST&
		jmp   short &INSTRUCT&_&CHAMP&
		endm




; ***********************************************************
;               moteur d'execution
;  excute une instruction pointe par PC : es:si et cl
; ***********************************************************


		align 2
_nexti:         xor  cl,4
                jne  _next
       		inc  si
_next:          dec  bp
		je   fin
_next0:         fetch_byte
		js   fixnext    ; no more needed on HP42S?
_next1:         and  ah,0fh
                mov  bx,ax
		add  bx,bx
		jmp  [tabopcode+bx]


NEXT            macro
		jmp  _next
		endm

NEXTI           macro
		jmp  _nexti
		endm



fixnext:	; segment suivant
		xor  bh,bh
		mov  bl,_regPC+2
		inc  bl
		mov  _regPC+2,bl
		add  bx,bx
		mov  bx,[_tab_seg+bx]
                or   bx,bx
                je   resetpc
                mov  es,bx
		and  si,07fffh
		jmp  short _next1
resetpc:       	; case es=0, jump to 0
                mov  _regPC+2,0
                xor  si,si
                xor  cl,cl
                jmp  short invalid   ;  !!!
                jmp  short _next




fin:
; fin de 256 instructions.
; test fin des cpt_instr instructions
;		mov  al,cpt_h
;		or   al,al
;		je   fin1
;		dec  al
;		mov  cpt_h,al
;		jmp  _next0
;fin1:
                xor  ax,ax
                jmp  short retour



invalid:        mov  ax,-1
                jmp  short retour
retouri:        incr_pc
retour:         push ax
		; convert regPC, si, cl en adr quartet 20 bits ah-di
		mov  ah,_regPC+2
		and  ah,0fh
		mov  al,cl
		mov  di,si
		shr  al,3   ; carry = bit 0
		rcl  di,1
		adc  ah,0        ; 02/01/97
		mov  word ptr _regPC,di
		mov  _regPC+2,ah

		irp  DEST,<A,B,C,D>
		mov  di,offset _reg&DEST&
		call maj_w
		endm
                mov  al,0fh
		and  _regD0+2,al
		and  _regD1+2,al

                IF FASTFLAGS
                mov  _Carry,0
                mov  _flagdec,0
                test ch,011h
                je   ff1
                mov  _Carry,0ffh
ff1:            test ch,080h
                je   ff2
                mov  _flagdec,0ffh
ff2:
                ENDIF

                ; setup stack pointer
                mov  al,_retstkptr
                shr  al,2
                mov  _retstkptr,al

		pop  ax
		ret




; ***********************************************************
;               point d'entree
; prototype C: int executeSaturn(int cpt);
; ***********************************************************


; point d'entre:
_executeSaturn  proc C far nb_instr:word, nb_ticks:word
		push si
		push di
	        mov  ax,nb_instr
		mov  bx,nb_ticks
                push bp
		call @executeSaturn
                pop  bp
                pop  di
		pop  si
		ret
_executeSaturn  endp



;_executeShndwn  proc C far nb_ticks: word
;		push si
;		push di
;		mov  bx,nb_ticks
;		add  cpt_ticks,bx
;		je   execShdn1
;               dec  cpt_ticks
;		call decr_timer
;execShdn1:	pop  di
;		pop  si
;		ret
;_executeShndwn  endp


@executeSaturn: mov  cpt_instr,ax
;;		add  cpt_ticks,bx
;;		cmp  ax,1
;;		jne  @exec1
;;		mov  cpt_ticks,0
@exec1:
                mov  al,tabloaded
                cmp  al,71
                je   @exec2
                call _init_tab
                mov  tabloaded,71
@exec2:
                ; setup stack
                mov  al,_retstkptr
                shl  al,2
                mov  _retstkptr,al

                ; setup PC
                mov  bl,_regPC+2
		and  bx,0fh
		mov  si,word ptr _regPC
		call bld_pointer       ; construit pointeur es:si,cl
                ; update A fields
		irp  DEST,<A,B,C,D>
		mov  di,offset _reg&DEST&
		maj_a
		endm
                ; setup bp as opcode counter
		mov  ax,cpt_instr
                mov  bp,ax
;;                xor  al,al
;;		mov  cpt_h,al


                ; setup flags
                IF FASTFLAGS
                xor  ch,ch
                test _Carry,0ffh
                je   fff1
                or   ch,1
fff1:           test _flagdec,0ffh
                je   fff2
                or   ch,080h
fff2:           test _Pointer,0fh
                je   fff3
                or   ch,2
fff3:
                ENDIF

		jmp  _next0






; ==============================================================

; decodage des instructions
; (dfinition de la table tab_opcode)

;		include sat_dec.inc

; ==============================================================









; ==================================================================

; dfinitions des instructions arithmtiques:
; x=x+y fs
; x=x-y fs
; x=y-x fs
; x=x+1 fs
; x=x-1 fs
; x=-x  fs
; x=-x-1 fs

;		include sat_arit.inc

; ==================================================================







; ==================================================================
; dfinitions des instructions de comparaison:
; ?x=y  fs
; ?x#y  fs
; ?x>y  fs
; ?x>=y fs
; ?x<y  fs
; ?x<=y fs
; ?x=0  fs
; ?x#0  fs

;		include sat_cmp.inc

; ==================================================================





; ==================================================================
; dfinitions des instructions logiques:
; x=x&y fs
; x=x!y fs
; xsl   fs
; xsr   fs
; x=0   fs
; x=y   fs
; xyex  fs
; Rn=x
; x=Rn
; xRnex

;		include sat_logi.inc

; ==================================================================





; ==================================================================
; dfinitions des instructions de saut:
; goc
; gonc
; goto
; gosub
; golong
; govlng
; gosubl
; gosbvl
; rtnsxm
; rtn
; rtnsc
; rtncc
; rti
; pc=a
; pc=c
; a=pc
; c=pc
; apcex
; cpcex
; pc=(a)
; pc=(c)

;		include sat_jmp.inc

; ==================================================================




; ==================================================================
; dfinitions des instructions de control:
; sethex
; setdec

; rstk=c
; c=rstk
; lchex
; lahex

; clrst
; c=st
; st=c
; cstex
; st=0  n
; st=1  n
; ?st=0 n
; ?st=1 n
; clrhst
; ?xx=0

; p=p+1
; p=p-1
; p=    n
; ?p=   n
; ?p#   n
; c+p+1
; c=p   n
; p=c   n
; cpex  n

; out=c
; out=cs
; a=in
; c=in
; uncnfg
; config
; c=id
; shutdn
; reset
; buscb
; buscc
; buscd
; sreq
; inton
; intoff
; rsi

; a=bit0 n
; a=bit1 n
; ?abit=0 n
; ?abit=1 n
; c=bit0 n
; c=bit1 n
; ?cbit=0 n
; ?cbit=1 n

; dx=(2)
; dx=(4)
; dx=(5)
; dx=dx+ n
; dx=dx- n
; dx=a
; dx=c
; adxex
; cdxex
; adxxs
; cdxxs

; a=a+con fs
; a=a-con fs
;		include sat_ctrl.inc

; ==================================================================



; ==================================================================
; dfinitions des instructions d'accs mmoire:
; a=dat0 fs
; c=dat0 fs
; a=dat1 fs
; c=dat1 fs
; dat0=a fs
; dat0=c fs
; dat1=a fs
; dat1=c fs

;		include sat_mem.inc

; ==================================================================


;		end

